# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'image_viewer.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(603, 833)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(self.centralwidget)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 569, 845))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_30 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_30.setObjectName("label_30")
        self.gridLayout_2.addWidget(self.label_30, 33, 0, 1, 1)
        self.label_fibsem_version = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_fibsem_version.setText("")
        self.label_fibsem_version.setObjectName("label_fibsem_version")
        self.gridLayout_2.addWidget(self.label_fibsem_version, 14, 1, 1, 2)
        self.line_5 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.line_5.setFont(font)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout_2.addWidget(self.line_5, 36, 0, 1, 3)
        self.line_2 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.line_2.setFont(font)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_2.addWidget(self.line_2, 21, 0, 1, 3)
        self.label_15 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_15.setFont(font)
        self.label_15.setObjectName("label_15")
        self.gridLayout_2.addWidget(self.label_15, 16, 0, 1, 1)
        self.label_name = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_name.setFont(font)
        self.label_name.setText("")
        self.label_name.setObjectName("label_name")
        self.gridLayout_2.addWidget(self.label_name, 1, 1, 1, 2)
        self.label_detector_mode = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_mode.setText("")
        self.label_detector_mode.setObjectName("label_detector_mode")
        self.gridLayout_2.addWidget(self.label_detector_mode, 19, 1, 1, 2)
        self.line_6 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        self.line_6.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.gridLayout_2.addWidget(self.line_6, 42, 0, 1, 3)
        self.label_11 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_11.setObjectName("label_11")
        self.gridLayout_2.addWidget(self.label_11, 11, 0, 1, 1)
        self.label_method = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_method.setText("")
        self.label_method.setObjectName("label_method")
        self.gridLayout_2.addWidget(self.label_method, 12, 1, 1, 2)
        self.label_20 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_20.setFont(font)
        self.label_20.setObjectName("label_20")
        self.gridLayout_2.addWidget(self.label_20, 22, 0, 1, 1)
        self.label_detector_brightness = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_brightness.setText("")
        self.label_detector_brightness.setObjectName("label_detector_brightness")
        self.gridLayout_2.addWidget(self.label_detector_brightness, 17, 1, 1, 2)
        self.label_14 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setItalic(False)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout_2.addWidget(self.label_14, 14, 0, 1, 1)
        self.label_21 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_21.setObjectName("label_21")
        self.gridLayout_2.addWidget(self.label_21, 23, 0, 1, 1)
        self.label_timestamp = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_timestamp.setFont(font)
        self.label_timestamp.setText("")
        self.label_timestamp.setObjectName("label_timestamp")
        self.gridLayout_2.addWidget(self.label_timestamp, 6, 1, 1, 2)
        self.label_application = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_application.setText("")
        self.label_application.setObjectName("label_application")
        self.gridLayout_2.addWidget(self.label_application, 13, 1, 1, 1)
        self.label_serial_no = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_serial_no.setText("")
        self.label_serial_no.setObjectName("label_serial_no")
        self.gridLayout_2.addWidget(self.label_serial_no, 34, 1, 1, 2)
        self.label_19 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_19.setObjectName("label_19")
        self.gridLayout_2.addWidget(self.label_19, 20, 0, 1, 1)
        self.label_shift = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_shift.setText("")
        self.label_shift.setObjectName("label_shift")
        self.gridLayout_2.addWidget(self.label_shift, 28, 1, 1, 2)
        self.label_app_version = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_app_version.setText("")
        self.label_app_version.setObjectName("label_app_version")
        self.gridLayout_2.addWidget(self.label_app_version, 13, 2, 1, 1)
        self.line_4 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.line_4.setFont(font)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout_2.addWidget(self.line_4, 9, 0, 1, 3)
        self.label_29 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_29.setObjectName("label_29")
        self.gridLayout_2.addWidget(self.label_29, 32, 0, 1, 1)
        self.label_16 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_16.setObjectName("label_16")
        self.gridLayout_2.addWidget(self.label_16, 17, 0, 1, 1)
        self.line_3 = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.line_3.setFont(font)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout_2.addWidget(self.line_3, 30, 0, 1, 3)
        self.label_31 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_31.setObjectName("label_31")
        self.gridLayout_2.addWidget(self.label_31, 34, 0, 1, 1)
        self.label_metadata_version = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_metadata_version.setFont(font)
        self.label_metadata_version.setText("")
        self.label_metadata_version.setObjectName("label_metadata_version")
        self.gridLayout_2.addWidget(self.label_metadata_version, 44, 1, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout_2.addWidget(self.label_10, 10, 0, 1, 1)
        self.line = QtWidgets.QFrame(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.line.setFont(font)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout_2.addWidget(self.line, 15, 0, 1, 3)
        self.label_2 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 3, 0, 1, 1)
        self.label_25 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_25.setObjectName("label_25")
        self.gridLayout_2.addWidget(self.label_25, 29, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 43, 0, 1, 3)
        self.label_6 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.gridLayout_2.addWidget(self.label_6, 5, 0, 1, 1)
        self.label_dwell_time = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_dwell_time.setText("")
        self.label_dwell_time.setObjectName("label_dwell_time")
        self.gridLayout_2.addWidget(self.label_dwell_time, 26, 1, 1, 2)
        self.label_software_version = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_software_version.setText("")
        self.label_software_version.setObjectName("label_software_version")
        self.gridLayout_2.addWidget(self.label_software_version, 35, 1, 1, 2)
        self.label_27 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_27.setObjectName("label_27")
        self.gridLayout_2.addWidget(self.label_27, 25, 0, 1, 1)
        self.label_manufacturer = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_manufacturer.setText("")
        self.label_manufacturer.setObjectName("label_manufacturer")
        self.gridLayout_2.addWidget(self.label_manufacturer, 32, 1, 1, 2)
        self.label_23 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_23.setObjectName("label_23")
        self.gridLayout_2.addWidget(self.label_23, 27, 0, 1, 1)
        self.label_28 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_28.setFont(font)
        self.label_28.setObjectName("label_28")
        self.gridLayout_2.addWidget(self.label_28, 31, 0, 1, 1)
        self.label_organisation = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_organisation.setText("")
        self.label_organisation.setObjectName("label_organisation")
        self.gridLayout_2.addWidget(self.label_organisation, 39, 1, 1, 2)
        self.label_beam_current = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_current.setText("")
        self.label_beam_current.setObjectName("label_beam_current")
        self.gridLayout_2.addWidget(self.label_beam_current, 23, 1, 1, 2)
        self.label_model = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_model.setText("")
        self.label_model.setObjectName("label_model")
        self.gridLayout_2.addWidget(self.label_model, 33, 1, 1, 2)
        self.label_stigmation = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_stigmation.setText("")
        self.label_stigmation.setObjectName("label_stigmation")
        self.gridLayout_2.addWidget(self.label_stigmation, 29, 1, 1, 2)
        self.label_working_distance = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_working_distance.setText("")
        self.label_working_distance.setObjectName("label_working_distance")
        self.gridLayout_2.addWidget(self.label_working_distance, 25, 1, 1, 2)
        self.label_32 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_32.setObjectName("label_32")
        self.gridLayout_2.addWidget(self.label_32, 35, 0, 1, 1)
        self.label_userpc = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_userpc.setText("")
        self.label_userpc.setObjectName("label_userpc")
        self.gridLayout_2.addWidget(self.label_userpc, 40, 1, 1, 2)
        self.label_18 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_18.setObjectName("label_18")
        self.gridLayout_2.addWidget(self.label_18, 19, 0, 1, 1)
        self.label_ID = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_ID.setText("")
        self.label_ID.setObjectName("label_ID")
        self.gridLayout_2.addWidget(self.label_ID, 11, 1, 1, 2)
        self.label_voltage = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_voltage.setText("")
        self.label_voltage.setObjectName("label_voltage")
        self.gridLayout_2.addWidget(self.label_voltage, 24, 1, 1, 2)
        self.label_37 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_37.setObjectName("label_37")
        self.gridLayout_2.addWidget(self.label_37, 41, 0, 1, 1)
        self.label_beamType = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_beamType.setFont(font)
        self.label_beamType.setText("")
        self.label_beamType.setObjectName("label_beamType")
        self.gridLayout_2.addWidget(self.label_beamType, 3, 1, 1, 2)
        self.label_33 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_33.setFont(font)
        self.label_33.setObjectName("label_33")
        self.gridLayout_2.addWidget(self.label_33, 37, 0, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_8.setFont(font)
        self.label_8.setObjectName("label_8")
        self.gridLayout_2.addWidget(self.label_8, 44, 0, 1, 1)
        self.label_26 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_26.setObjectName("label_26")
        self.gridLayout_2.addWidget(self.label_26, 24, 0, 1, 1)
        self.label_detector_contrast = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_contrast.setText("")
        self.label_detector_contrast.setObjectName("label_detector_contrast")
        self.gridLayout_2.addWidget(self.label_detector_contrast, 18, 1, 1, 2)
        self.label_username = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_username.setText("")
        self.label_username.setObjectName("label_username")
        self.gridLayout_2.addWidget(self.label_username, 38, 1, 1, 2)
        self.label_12 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_12.setObjectName("label_12")
        self.gridLayout_2.addWidget(self.label_12, 12, 0, 1, 1)
        self.label_hfw = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_hfw.setFont(font)
        self.label_hfw.setText("")
        self.label_hfw.setObjectName("label_hfw")
        self.gridLayout_2.addWidget(self.label_hfw, 5, 1, 1, 2)
        self.label_34 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_34.setObjectName("label_34")
        self.gridLayout_2.addWidget(self.label_34, 38, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout_2.addWidget(self.label_5, 4, 0, 1, 1)
        self.label_detector_type = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_type.setText("")
        self.label_detector_type.setObjectName("label_detector_type")
        self.gridLayout_2.addWidget(self.label_detector_type, 20, 1, 1, 2)
        self.label_17 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_17.setObjectName("label_17")
        self.gridLayout_2.addWidget(self.label_17, 18, 0, 1, 1)
        self.label_38 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label_38.setFont(font)
        self.label_38.setObjectName("label_38")
        self.gridLayout_2.addWidget(self.label_38, 7, 0, 1, 1)
        self.label_35 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_35.setObjectName("label_35")
        self.gridLayout_2.addWidget(self.label_35, 39, 0, 1, 1)
        self.label_24 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_24.setObjectName("label_24")
        self.gridLayout_2.addWidget(self.label_24, 28, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.gridLayout_2.addWidget(self.label_7, 1, 0, 1, 1)
        self.label_resolution = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_resolution.setFont(font)
        self.label_resolution.setText("")
        self.label_resolution.setObjectName("label_resolution")
        self.gridLayout_2.addWidget(self.label_resolution, 4, 1, 1, 2)
        self.label_13 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_13.setObjectName("label_13")
        self.gridLayout_2.addWidget(self.label_13, 13, 0, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(False)
        font.setWeight(50)
        self.label_9.setFont(font)
        self.label_9.setObjectName("label_9")
        self.gridLayout_2.addWidget(self.label_9, 6, 0, 1, 1)
        self.label_36 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_36.setObjectName("label_36")
        self.gridLayout_2.addWidget(self.label_36, 40, 0, 1, 1)
        self.label_scan_rotation = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_scan_rotation.setText("")
        self.label_scan_rotation.setObjectName("label_scan_rotation")
        self.gridLayout_2.addWidget(self.label_scan_rotation, 27, 1, 1, 2)
        self.label_22 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_22.setObjectName("label_22")
        self.gridLayout_2.addWidget(self.label_22, 26, 0, 1, 1)
        self.label_email = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_email.setText("")
        self.label_email.setObjectName("label_email")
        self.gridLayout_2.addWidget(self.label_email, 41, 1, 1, 2)
        self.label_position = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_position.setText("")
        self.label_position.setObjectName("label_position")
        self.gridLayout_2.addWidget(self.label_position, 8, 0, 1, 3)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 4, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.centralwidget)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 2, 0, 1, 1)
        self.pushButton_export = QtWidgets.QPushButton(self.centralwidget)
        self.pushButton_export.setObjectName("pushButton_export")
        self.gridLayout.addWidget(self.pushButton_export, 3, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 603, 20))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_30.setText(_translate("MainWindow", "Model"))
        self.label_15.setText(_translate("MainWindow", "Detector Settings"))
        self.label_11.setText(_translate("MainWindow", "Experiment ID"))
        self.label_20.setText(_translate("MainWindow", "Beam Settings"))
        self.label_14.setText(_translate("MainWindow", "OpenFIBSEM Version"))
        self.label_21.setText(_translate("MainWindow", "Beam Current"))
        self.label_19.setText(_translate("MainWindow", "Type"))
        self.label_29.setText(_translate("MainWindow", "Manufacturer"))
        self.label_16.setText(_translate("MainWindow", "Brightness"))
        self.label_31.setText(_translate("MainWindow", "Serial No."))
        self.label_10.setText(_translate("MainWindow", "Experiment"))
        self.label_2.setText(_translate("MainWindow", "Beam Type"))
        self.label_25.setText(_translate("MainWindow", "Stigmation"))
        self.label_6.setText(_translate("MainWindow", "Horizontal Field Width (um)"))
        self.label_27.setText(_translate("MainWindow", "Working Distance"))
        self.label_23.setText(_translate("MainWindow", "Scan Rotation"))
        self.label_28.setText(_translate("MainWindow", "System"))
        self.label_32.setText(_translate("MainWindow", "Software Version"))
        self.label_18.setText(_translate("MainWindow", "Mode"))
        self.label_37.setText(_translate("MainWindow", "Email"))
        self.label_33.setText(_translate("MainWindow", "User"))
        self.label_8.setText(_translate("MainWindow", "Image Metadata Version"))
        self.label_26.setText(_translate("MainWindow", "Voltage"))
        self.label_12.setText(_translate("MainWindow", "Method"))
        self.label_34.setText(_translate("MainWindow", "Name"))
        self.label_5.setText(_translate("MainWindow", "Resolution"))
        self.label_17.setText(_translate("MainWindow", "Contrast"))
        self.label_38.setText(_translate("MainWindow", "Stage Position"))
        self.label_35.setText(_translate("MainWindow", "Organisation"))
        self.label_24.setText(_translate("MainWindow", "Shift"))
        self.label_7.setText(_translate("MainWindow", "Filename"))
        self.label_13.setText(_translate("MainWindow", "Application (Version)"))
        self.label_9.setText(_translate("MainWindow", "Time Stamp "))
        self.label_36.setText(_translate("MainWindow", "Computer"))
        self.label_22.setText(_translate("MainWindow", "Dwell Time"))
        self.label_3.setText(_translate("MainWindow", "Drag and drop images into the viewer"))
        self.label.setText(_translate("MainWindow", "OpenFIBSEM Image Viewer"))
        self.label_4.setText(_translate("MainWindow", "Drag and drop of folders is currently unsupported"))
        self.pushButton_export.setText(_translate("MainWindow", "Export Image with Metadata"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemLabellingUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(400, 469)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.comboBox_model_class_index = QtWidgets.QComboBox(Dialog)
        self.comboBox_model_class_index.setObjectName("comboBox_model_class_index")
        self.gridLayout.addWidget(self.comboBox_model_class_index, 4, 1, 1, 1)
        self.checkBox_model_assist = QtWidgets.QCheckBox(Dialog)
        self.checkBox_model_assist.setObjectName("checkBox_model_assist")
        self.gridLayout.addWidget(self.checkBox_model_assist, 2, 0, 1, 1)
        self.label_model_info = QtWidgets.QLabel(Dialog)
        self.label_model_info.setObjectName("label_model_info")
        self.gridLayout.addWidget(self.label_model_info, 2, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 13, 0, 1, 2)
        self.label_instructions = QtWidgets.QLabel(Dialog)
        self.label_instructions.setWordWrap(True)
        self.label_instructions.setObjectName("label_instructions")
        self.gridLayout.addWidget(self.label_instructions, 11, 0, 1, 2)
        self.pushButton_model_clear = QtWidgets.QPushButton(Dialog)
        self.pushButton_model_clear.setObjectName("pushButton_model_clear")
        self.gridLayout.addWidget(self.pushButton_model_clear, 5, 1, 1, 1)
        self.pushButton_model_confirm = QtWidgets.QPushButton(Dialog)
        self.pushButton_model_confirm.setObjectName("pushButton_model_confirm")
        self.gridLayout.addWidget(self.pushButton_model_confirm, 5, 0, 1, 1)
        self.label_title = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 2)
        self.tabWidget = QtWidgets.QTabWidget(Dialog)
        self.tabWidget.setMinimumSize(QtCore.QSize(0, 0))
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.tab)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_data_file_ext = QtWidgets.QLabel(self.tab)
        self.label_data_file_ext.setObjectName("label_data_file_ext")
        self.gridLayout_2.addWidget(self.label_data_file_ext, 3, 0, 1, 1)
        self.label_labels_path = QtWidgets.QLabel(self.tab)
        self.label_labels_path.setObjectName("label_labels_path")
        self.gridLayout_2.addWidget(self.label_labels_path, 1, 0, 1, 1)
        self.checkBox_save_rgb = QtWidgets.QCheckBox(self.tab)
        self.checkBox_save_rgb.setChecked(False)
        self.checkBox_save_rgb.setObjectName("checkBox_save_rgb")
        self.gridLayout_2.addWidget(self.checkBox_save_rgb, 6, 1, 1, 1)
        self.pushButton_labels_path = QtWidgets.QToolButton(self.tab)
        self.pushButton_labels_path.setObjectName("pushButton_labels_path")
        self.gridLayout_2.addWidget(self.pushButton_labels_path, 1, 2, 1, 1)
        self.lineEdit_labels_path = QtWidgets.QLineEdit(self.tab)
        self.lineEdit_labels_path.setObjectName("lineEdit_labels_path")
        self.gridLayout_2.addWidget(self.lineEdit_labels_path, 1, 1, 1, 1)
        self.pushButton_load_data = QtWidgets.QPushButton(self.tab)
        self.pushButton_load_data.setObjectName("pushButton_load_data")
        self.gridLayout_2.addWidget(self.pushButton_load_data, 4, 0, 1, 3)
        self.label_data_path = QtWidgets.QLabel(self.tab)
        self.label_data_path.setObjectName("label_data_path")
        self.gridLayout_2.addWidget(self.label_data_path, 0, 0, 1, 1)
        self.lineEdit_data_path = QtWidgets.QLineEdit(self.tab)
        self.lineEdit_data_path.setObjectName("lineEdit_data_path")
        self.gridLayout_2.addWidget(self.lineEdit_data_path, 0, 1, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem1, 5, 0, 1, 3)
        self.pushButton_data_path = QtWidgets.QToolButton(self.tab)
        self.pushButton_data_path.setObjectName("pushButton_data_path")
        self.gridLayout_2.addWidget(self.pushButton_data_path, 0, 2, 1, 1)
        self.comboBox_data_file_ext = QtWidgets.QComboBox(self.tab)
        self.comboBox_data_file_ext.setObjectName("comboBox_data_file_ext")
        self.gridLayout_2.addWidget(self.comboBox_data_file_ext, 3, 1, 1, 2)
        self.checkBox_autosave = QtWidgets.QCheckBox(self.tab)
        self.checkBox_autosave.setChecked(True)
        self.checkBox_autosave.setObjectName("checkBox_autosave")
        self.gridLayout_2.addWidget(self.checkBox_autosave, 6, 0, 1, 1)
        self.label_data_config = QtWidgets.QLabel(self.tab)
        self.label_data_config.setObjectName("label_data_config")
        self.gridLayout_2.addWidget(self.label_data_config, 2, 0, 1, 1)
        self.lineEdit_data_config = QtWidgets.QLineEdit(self.tab)
        self.lineEdit_data_config.setObjectName("lineEdit_data_config")
        self.gridLayout_2.addWidget(self.lineEdit_data_config, 2, 1, 1, 1)
        self.pushButton_data_config = QtWidgets.QToolButton(self.tab)
        self.pushButton_data_config.setObjectName("pushButton_data_config")
        self.gridLayout_2.addWidget(self.pushButton_data_config, 2, 2, 1, 1)
        self.tabWidget.addTab(self.tab, "")
        self.gridLayout.addWidget(self.tabWidget, 1, 0, 1, 2)
        self.label_model_class_index = QtWidgets.QLabel(Dialog)
        self.label_model_class_index.setObjectName("label_model_class_index")
        self.gridLayout.addWidget(self.label_model_class_index, 4, 0, 1, 1)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.tabWidget, self.lineEdit_data_path)
        Dialog.setTabOrder(self.lineEdit_data_path, self.pushButton_data_path)
        Dialog.setTabOrder(self.pushButton_data_path, self.lineEdit_labels_path)
        Dialog.setTabOrder(self.lineEdit_labels_path, self.pushButton_labels_path)
        Dialog.setTabOrder(self.pushButton_labels_path, self.pushButton_load_data)
        Dialog.setTabOrder(self.pushButton_load_data, self.checkBox_model_assist)
        Dialog.setTabOrder(self.checkBox_model_assist, self.comboBox_model_class_index)
        Dialog.setTabOrder(self.comboBox_model_class_index, self.pushButton_model_confirm)
        Dialog.setTabOrder(self.pushButton_model_confirm, self.pushButton_model_clear)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.checkBox_model_assist.setText(_translate("Dialog", "Model Assisted"))
        self.label_model_info.setText(_translate("Dialog", "No Model"))
        self.label_instructions.setText(_translate("Dialog", "Instructions"))
        self.pushButton_model_clear.setText(_translate("Dialog", "Clear"))
        self.pushButton_model_confirm.setText(_translate("Dialog", "Confirm"))
        self.label_title.setText(_translate("Dialog", "OpenFIBSEM Labelling"))
        self.label_data_file_ext.setText(_translate("Dialog", "File Extension"))
        self.label_labels_path.setText(_translate("Dialog", "Label Path"))
        self.checkBox_save_rgb.setText(_translate("Dialog", "Save RGB Masks"))
        self.pushButton_labels_path.setText(_translate("Dialog", "..."))
        self.pushButton_load_data.setText(_translate("Dialog", "Load Data"))
        self.label_data_path.setText(_translate("Dialog", "Image Path"))
        self.pushButton_data_path.setText(_translate("Dialog", "..."))
        self.checkBox_autosave.setText(_translate("Dialog", "AutoSave"))
        self.label_data_config.setText(_translate("Dialog", "Class Config"))
        self.pushButton_data_config.setText(_translate("Dialog", "..."))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("Dialog", "Data"))
        self.label_model_class_index.setText(_translate("Dialog", "Class Map"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemEmbeddedDetectionWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(375, 465)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.label_labelling_class_index = QtWidgets.QLabel(Form)
        self.label_labelling_class_index.setObjectName("label_labelling_class_index")
        self.gridLayout.addWidget(self.label_labelling_class_index, 8, 0, 1, 1)
        self.label_labelling_instructions = QtWidgets.QLabel(Form)
        self.label_labelling_instructions.setWordWrap(True)
        self.label_labelling_instructions.setObjectName("label_labelling_instructions")
        self.gridLayout.addWidget(self.label_labelling_instructions, 10, 0, 1, 2)
        self.label_info = QtWidgets.QLabel(Form)
        self.label_info.setMinimumSize(QtCore.QSize(0, 0))
        self.label_info.setText("")
        self.label_info.setWordWrap(True)
        self.label_info.setObjectName("label_info")
        self.gridLayout.addWidget(self.label_info, 2, 0, 1, 2)
        self.pushButton_labelling_cancel = QtWidgets.QPushButton(Form)
        self.pushButton_labelling_cancel.setObjectName("pushButton_labelling_cancel")
        self.gridLayout.addWidget(self.pushButton_labelling_cancel, 9, 1, 1, 1)
        self.comboBox_labelling_class_index = QtWidgets.QComboBox(Form)
        self.comboBox_labelling_class_index.setObjectName("comboBox_labelling_class_index")
        self.gridLayout.addWidget(self.comboBox_labelling_class_index, 8, 1, 1, 1)
        self.label_instructions = QtWidgets.QLabel(Form)
        self.label_instructions.setWordWrap(True)
        self.label_instructions.setObjectName("label_instructions")
        self.gridLayout.addWidget(self.label_instructions, 3, 0, 1, 2)
        self.pushButton_enable_labelling = QtWidgets.QPushButton(Form)
        self.pushButton_enable_labelling.setObjectName("pushButton_enable_labelling")
        self.gridLayout.addWidget(self.pushButton_enable_labelling, 6, 0, 1, 2)
        self.label_model = QtWidgets.QLabel(Form)
        self.label_model.setObjectName("label_model")
        self.gridLayout.addWidget(self.label_model, 1, 0, 1, 2)
        self.pushButton_labelling_confirm = QtWidgets.QPushButton(Form)
        self.pushButton_labelling_confirm.setObjectName("pushButton_labelling_confirm")
        self.gridLayout.addWidget(self.pushButton_labelling_confirm, 9, 0, 1, 1)
        self.label_labelling_model = QtWidgets.QLabel(Form)
        self.label_labelling_model.setText("")
        self.label_labelling_model.setObjectName("label_labelling_model")
        self.gridLayout.addWidget(self.label_labelling_model, 7, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 11, 0, 1, 2)
        self.pushButton_continue = QtWidgets.QPushButton(Form)
        self.pushButton_continue.setObjectName("pushButton_continue")
        self.gridLayout.addWidget(self.pushButton_continue, 4, 0, 1, 2)
        self.checkBox_labelling_model_assist = QtWidgets.QCheckBox(Form)
        self.checkBox_labelling_model_assist.setObjectName("checkBox_labelling_model_assist")
        self.gridLayout.addWidget(self.checkBox_labelling_model_assist, 7, 0, 1, 1)
        self.label_labelling_header = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_labelling_header.setFont(font)
        self.label_labelling_header.setObjectName("label_labelling_header")
        self.gridLayout.addWidget(self.label_labelling_header, 5, 0, 1, 2)
        self.label_title = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(9)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_labelling_class_index.setText(_translate("Form", "Class Index"))
        self.label_labelling_instructions.setText(_translate("Form", "Use the paint tools to label the images."))
        self.pushButton_labelling_cancel.setText(_translate("Form", "Cancel"))
        self.label_instructions.setText(_translate("Form", "Drag to move the features, when finished press confirm."))
        self.pushButton_enable_labelling.setText(_translate("Form", "Enable Labelling"))
        self.label_model.setText(_translate("Form", "Model:"))
        self.pushButton_labelling_confirm.setText(_translate("Form", "Confirm"))
        self.pushButton_continue.setText(_translate("Form", "Confirm"))
        self.checkBox_labelling_model_assist.setText(_translate("Form", "Segment Anything Assist"))
        self.label_labelling_header.setText(_translate("Form", "Labelling"))
        self.label_title.setText(_translate("Form", "Feature Detection"))



import numpy as np
import torch
from transformers import SamModel, SamProcessor


class SamModelWrapper:
    def __init__(self, checkpoint: str = "facebook/sam-vit-base", device: str = None):
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device
        model = SamModel.from_pretrained(checkpoint).to(self.device)
        processor = SamProcessor.from_pretrained(checkpoint)

        self.checkpoint = checkpoint
        self.model = model
        self.processor = processor

    def __call__(
        self,
        image: np.ndarray,
        points: list[list[list[int]]] = None,
        labels: list[list[bool]] = None,
        input_masks=None,
        multimask_output: bool = False,
    ):
        inputs = self.processor(
            image,
            input_points=points,
            input_labels=labels,
            return_tensors="pt",
            input_masks=input_masks,
            multimask_output=multimask_output,
        ).to(self.device)
        # TODO: multi-mask output doesn't seem to work?
        with torch.no_grad():
            outputs = self.model(**inputs)
        return outputs, inputs

    # to mimic the functionality of sam predictor
    def predict(
        self,
        image,
        points: list[list[list[int]]] = None,
        labels: list[list[bool]] = None,
        input_masks: np.ndarray = None,
        multimask_output: bool = False,
    ):
        outputs, inputs = self(image, points, labels, input_masks, multimask_output)
        masks = self.processor.image_processor.post_process_masks(
            outputs.pred_masks.cpu(),
            inputs["original_sizes"].cpu(),
            inputs["reshaped_input_sizes"].cpu(),
        )
        scores = outputs.iou_scores
        logits = outputs.pred_masks

        # get best mask
        idx = np.argmax(scores.detach().cpu().numpy())
        mask = masks[0][0][idx].detach().cpu().numpy()
        score = scores[0][0][idx].detach().cpu().numpy()
        logits = logits[0][0][idx].detach().cpu().unsqueeze(0).numpy()

        return mask, score, logits

    def __repr__(self):
        return f"SamModelWrapper({self.checkpoint}, {self.device})"



import pytest
from fibsem import utils, acquire, alignment
import random
import numpy as np

def test_align_from_crosscorrelation():
    
    microscope, settings = utils.setup_session(debug=False)

    # create random images
    ref_image = acquire.acquire_image(microscope, settings.image)
    new_image = acquire.acquire_image(microscope, settings.image)

    ref_image.data[:] = 0
    new_image.data[:] = 0

    # crop a random square out
    w = h = 150
    x = random.randint(0, ref_image.data.shape[1] - 2*w)
    y = random.randint(0, ref_image.data.shape[0] - 2*h)
    ref_image.data[y:y+h, x:x+w]  = 255

    # new image should be offset by 250 pixels in x and y
    offset = random.randint(-50, 50)
    new_image.data[y+offset:y+h+offset, x+offset:x+w+offset] = 255

    dx, dy, xcorr = alignment.shift_from_crosscorrelation(
        ref_image, new_image, lowpass=50, highpass=4, sigma=5, use_rect_mask=True
    )    

    # write a test case with pytest for this case
    # test that the shift is within 1 pixel of the offset

    pixel_size = ref_image.metadata.pixel_size.x
    assert np.isclose(dx, offset*pixel_size, atol=pixel_size), f"dx: {dx}, offset: {offset*pixel_size}"
    assert np.isclose(dy, offset*pixel_size, atol=pixel_size), f"dy: {dy}, offset: {offset*pixel_size}"


import datetime
import glob
import json
import logging
import os
import numpy as np
import time
from pathlib import Path
import sys

import yaml

from PIL import Image
from fibsem.structures import (
    BeamType,
    MicroscopeSettings,
    ImageSettings,
    SystemSettings,
    FibsemImage,
    FibsemMillingSettings,
)
from fibsem import config as cfg
from fibsem.microscope import FibsemMicroscope


def current_timestamp():
    """Returns current time in a specific string format

    Returns:
        String: Current time
    """
    return datetime.datetime.fromtimestamp(time.time()).strftime("%Y-%m-%d-%I-%M-%S%p") #PM/AM doesnt work?


def current_timestamp_v2():
    """Returns current time in a specific string format

    Returns:
        String: Current time
    """
    return str(time.time()).replace(".", "_")

def _format_time_seconds(seconds: float) -> str:
    """Format a time delta in seconds to proper string format."""
    return str(datetime.timedelta(seconds=seconds)).split(".")[0]


def make_logging_directory(path: Path = None, name="run"):
    """
    Create a logging directory with the specified name at the specified file path. 
    If no path is given, it creates the directory at the default base path.

    Args:
        path (Path, optional): The file path to create the logging directory at. If None, default base path is used. 
        name (str, optional): The name of the logging directory to create. Default is "run".

    Returns:
        str: The file path to the created logging directory.
        """
    
    if path is None:
        path = os.path.join(cfg.BASE_PATH, "log")
    directory = os.path.join(path, name)
    os.makedirs(directory, exist_ok=True)
    return directory

# TODO: better logs: https://www.toptal.com/python/in-depth-python-logging
# https://stackoverflow.com/questions/61483056/save-logging-debug-and-show-only-logging-info-python
def configure_logging(path: Path = "", log_filename="logfile", log_level=logging.DEBUG, _DEBUG: bool = False):
    """Log to the terminal and to file simultaneously."""
    logfile = os.path.join(path, f"{log_filename}.log")

    file_handler = logging.FileHandler(logfile)
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.INFO if _DEBUG is False else logging.DEBUG)

    logging.basicConfig(
        format="%(asctime)s — %(name)s — %(levelname)s — %(funcName)s:%(lineno)d — %(message)s",
        level=log_level,
        # Multiple handlers can be added to your logging configuration.
        # By default log messages are appended to the file if it exists already
        handlers=[file_handler, stream_handler],
        force=True,
    )

    return logfile


def load_yaml(fname: Path) -> dict:
    """load yaml file

    Args:
        fname (Path): yaml file path

    Returns:
        dict: Items in yaml
    """
    with open(fname, "r") as f:
        config = yaml.safe_load(f)

    return config


def save_yaml(path: Path, data: dict) -> None:
    """Saves a python dictionary object to a yaml file

    Args:
        path (Path): path location to save yaml file
        data (dict): dictionary object
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)
    path = Path(path).with_suffix(".yaml")
    with open(path, "w") as f:
        yaml.dump(data, f, indent=4)


def create_gif(path: Path, search: str, gif_fname: str, loop: int = 0) -> None:
    """Creates a GIF from a set of images. Images must be in same folder

    Args:
        path (Path): Path to images folder
        search (str): search name
        gif_fname (str): name to save gif file
        loop (int, optional): _description_. Defaults to 0.
    """
    filenames = glob.glob(os.path.join(path, search))

    imgs = [Image.fromarray(FibsemImage.load(fname).data) for fname in filenames]

    print(f"{len(filenames)} images added to gif.")
    imgs[0].save(
        os.path.join(path, f"{gif_fname}.gif"),
        save_all=True,
        append_images=imgs[1:],
        loop=loop,
    )

VALID_THERMO_FISHER = ["Thermo", "Thermo Fisher Scientific", "Thermo Fisher Scientific"]
VALID_TESCAN = ["Tescan", "TESCAN" ]

def setup_session(
    session_path: Path = None,
    config_path: Path = None,
    protocol_path: Path = None,
    setup_logging: bool = True,
    ip_address: str = None,
    manufacturer: str = None,
    debug: bool = False,
) -> tuple[FibsemMicroscope, MicroscopeSettings]:
    """Setup microscope session

    Args:
        session_path (Path): path to logging directory
        config_path (Path): path to config directory 
        protocol_path (Path): path to protocol file

    Returns:
        tuple: microscope, settings
    """

    # load settings
    settings = load_microscope_configuration(config_path, protocol_path)

    # create session directories
    session = f'{settings.protocol.get("name", "openfibsem")}_{current_timestamp()}'
    if protocol_path is None:
        protocol_path = os.getcwd()

    # configure paths
    if session_path is None:
        session_path = cfg.LOG_PATH
    os.makedirs(session_path, exist_ok=True)

    # configure logging
    if setup_logging:
        configure_logging(session_path, _DEBUG=debug)

    # connect to microscope
    import fibsem.microscope as fibsem_microscope

    # cheap overloading
    if ip_address:
        settings.system.info.ip_address = ip_address
    
    if manufacturer:
        settings.system.info.manufacturer = manufacturer

    manufacturer = settings.system.info.manufacturer
    ip_address = settings.system.info.ip_address

    if manufacturer in VALID_THERMO_FISHER:
        microscope = fibsem_microscope.ThermoMicroscope(settings.system)
        microscope.connect_to_microscope(
            ip_address=ip_address, port=7520
        )

    elif manufacturer in VALID_TESCAN:
        microscope = fibsem_microscope.TescanMicroscope(settings.system)
        microscope.connect_to_microscope(
            ip_address=ip_address, port=8300
        )
    
    elif manufacturer == "Demo":
        microscope = fibsem_microscope.DemoMicroscope(settings.system)
        microscope.connect_to_microscope(ip_address, port=7520)

    else:
        raise NotImplementedError(f"Manufacturer {manufacturer} not supported.")
    
    # set default image_settings path
    settings.image.path = session_path

    logging.info(f"Finished setup for session: {session}")

    return microscope, settings


def load_microscope_configuration(
    config_path: Path = None, protocol_path: Path = None
) -> MicroscopeSettings:
    """Load microscope settings from configuration files

    Args:
        config_path (Path, optional): path to config directory. Defaults to None.
        protocol_path (Path, optional): path to protocol file. Defaults to None.

    Returns:
        MicroscopeSettings: microscope settings
    """
    if config_path is None:
        from fibsem.config import DEFAULT_CONFIGURATION_PATH
        config_path = DEFAULT_CONFIGURATION_PATH
    
    # load config
    config = load_yaml(os.path.join(config_path))

    # load protocol
    protocol = load_protocol(protocol_path)

    # create settings
    settings = MicroscopeSettings.from_dict(config, protocol=protocol)

    return settings

def load_protocol(protocol_path: Path = None) -> dict:
    """Load the protocol file from yaml

    Args:
        protocol_path (Path, optional): path to protocol file. Defaults to None.

    Returns:
        dict: protocol dictionary
    """
    if protocol_path is not None:
        protocol = load_yaml(protocol_path)
    else:
        protocol = {"name": "demo"}

    #protocol = _format_dictionary(protocol)

    return protocol


def _format_dictionary(dictionary: dict) -> dict:
    """Recursively traverse dictionary and covert all numeric values to flaot.

    Parameters
    ----------
    dictionary : dict
        Any arbitrarily structured python dictionary.

    Returns
    -------
    dictionary
        The input dictionary, with all numeric values converted to float type.
    """
    for key, item in dictionary.items():
        if isinstance(item, dict):
            _format_dictionary(item)
        elif isinstance(item, list):
            dictionary[key] = [
                _format_dictionary(i)
                for i in item
                if isinstance(i, list) or isinstance(i, dict)
            ]
        else:
            if item is not None:
                try:
                    dictionary[key] = float(dictionary[key])
                except ValueError:
                    pass
    return dictionary

def get_params(main_str: str) -> list:
    """Helper function to access relevant metadata parameters from sub field

    Args:
        main_str (str): Sub string of relevant metadata

    Returns:
        list: Parameters covered by metadata
    """
    cats = []
    cat_str = ""

    i = main_str.find("\n")
    i += 1
    while i < len(main_str):

        if main_str[i] == "=":
            cats.append(cat_str)
            cat_str = ""
            i += main_str[i:].find("\n")
        else:
            cat_str += main_str[i]

        i += 1
    return cats


def _get_position(name: str):
    
    from fibsem import config as cfg
    from fibsem.structures import FibsemStagePosition
    import os

    ddict = load_yaml(fname=os.path.join(cfg.CONFIG_PATH, "positions.yaml"))
    # get position from save positions?
    for d in ddict:
        if d["name"] == name:
            return FibsemStagePosition.from_dict(d)
    return None

def _get_positions(fname: str = None) -> list[str]:    
    
    from fibsem import config as cfg
    from fibsem.structures import FibsemStagePosition
    import os

    if fname is None:
        fname = os.path.join(cfg.CONFIG_PATH, "positions.yaml")

    ddict = load_yaml(fname=fname)

    return [d["name"] for d in ddict]


import yaml



def save_positions(positions: list, path: str = None, overwrite: bool = False) -> None:
    """save the list of positions to file"""

    from fibsem.structures import FibsemStagePosition
    from fibsem import config as cfg

    # convert single position to list
    if not isinstance(positions, list):
        positions = [positions]

    # default path
    if path is None:
        path = cfg.POSITION_PATH

    # get existing positions    
    pdict = []
    if not overwrite:
        pdict = load_yaml(fname=path)

    
    # append new positions
    for position in positions:
        pdict.append(position.to_dict())
    
    # save
    save_yaml(path, pdict)
    

def _display_metadata(img: FibsemImage, timezone: str = 'Australia/Sydney', show: bool = True):
    import pytz
    from matplotlib_scalebar.scalebar import ScaleBar
    import matplotlib.pyplot as plt
    import fibsem.constants as constants
    fig, ax = plt.subplots()

    # Display the image
    ax.imshow(img.data, cmap='gray')
    image_height, image_width = img.data.shape
    # Hide axis
    ax.axis('off')
    ax.set_xlim(0, image_width)  # Set the width of the image
    ax.set_ylim(0, image_height)  # Set the height of the image
    # Create a list to store metadata lines
    if img.metadata.image_settings.beam_type == BeamType.ELECTRON:
        metadata_lines = "Electron Beam \n"
    else: 
        metadata_lines = "Ion Beam \n"

    # add metadata lines
    if img.metadata.image_settings.beam_type == BeamType.ELECTRON and img.metadata.microscope_state.electron_beam.voltage is not None:
        metadata_lines += f'{img.metadata.microscope_state.electron_beam.voltage * constants.SI_TO_KILO} kV  |  '
    elif img.metadata.image_settings.beam_type == BeamType.ION and img.metadata.microscope_state.ion_beam.voltage is not None:
        metadata_lines += f'{img.metadata.microscope_state.ion_beam.voltage * constants.SI_TO_KILO} kV  |  '
    else:
        metadata_lines += 'Voltage: Unknown  |  '
    metadata_lines += (f'HFW: {img.metadata.image_settings.hfw * constants.SI_TO_MICRO} μm  | ')
    metadata_lines += (f'{img.metadata.image_settings.resolution[0]} x {img.metadata.image_settings.resolution[1]}  |  ')

    desired_timezone = pytz.timezone(timezone)  
    timestamp = img.metadata.microscope_state.timestamp

    if isinstance(timestamp, str):
        timestamp_format = "%m/%d/%Y %H:%M:%S"
        timestamp = datetime.datetime.strptime(img.metadata.microscope_state.timestamp, timestamp_format)

    if isinstance(timestamp, (int,float)):
        timestamp_str = datetime.datetime.fromtimestamp(timestamp, tz=desired_timezone).strftime('%Y-%m-%d %I:%M %p')    
    
    if isinstance(timestamp, datetime.datetime):
        timestamp_str = timestamp.astimezone(desired_timezone).strftime('%Y-%m-%d %I:%M %p')         

    metadata_lines += (f"{timestamp_str}")

    # add empty char to second line to fill up space
    _line2 = metadata_lines.split("\n")[1]
    metadata_lines += " " * (70 - len(_line2))

    metadata_rect = plt.text(
            0.01, 0.03, metadata_lines,
            transform=ax.transAxes,
            fontsize=10,
            color='white',
            bbox=dict(facecolor='black', alpha=0.7),
            ha='left',
        )

    metadata_rect.set_clip_box(dict(width=1.0))
    scale = (img.metadata.image_settings.hfw * constants.SI_TO_MICRO) / img.data.shape[1]
    
    # transparent background
    scalebar = ScaleBar(scale, "um", 
        color="black", box_color="white", box_alpha=0.3) 

    plt.gca().add_artist(scalebar)

    if show:
        plt.show()

    return fig

# TODO: re-think this, dont like the pop ups
def _register_metadata(microscope: FibsemMicroscope, application_software: str, application_software_version: str, experiment_name: str, experiment_method: str) -> None:
    from fibsem.structures import FibsemUser, FibsemExperiment
    import fibsem

    user = FibsemUser.from_environment()

    experiment = FibsemExperiment(
        id = experiment_name,
        method=experiment_method,
        application=application_software, 
        fibsem_version=fibsem.__version__,
        application_version=application_software_version,
    )
    microscope.user = user
    microscope.experiment = experiment










################## MIGRATE FROM OLD AUTOLIFTOUT
# import os
# from pathlib import Path

# import numpy as np
# from fibsem.structures import ImageSettings

# def plot_two_images(img1, img2) -> None:
#     import matplotlib.pyplot as plt
#     from fibsem.structures import Point

#     c = Point(img1.data.shape[1] // 2, img1.data.shape[0] // 2)

#     fig, ax = plt.subplots(1, 2, figsize=(30, 30))
#     ax[0].imshow(img1.data, cmap="gray")
#     ax[0].plot(c.x, c.y, "y+", ms=50, markeredgewidth=2)
#     ax[1].imshow(img2.data, cmap="gray")
#     ax[1].plot(c.x, c.y, "y+", ms=50, markeredgewidth=2)
#     plt.show()


# def take_reference_images_and_plot(microscope, image_settings: ImageSettings):
#     from pprint import pprint

#     from fibsem import acquire

#     eb_image, ib_image = acquire.take_reference_images(microscope, image_settings)
#     plot_two_images(eb_image, ib_image)

#     return eb_image, ib_image


# # cross correlate
# def crosscorrelate_and_plot(
#     ref_image,
#     new_image,
#     rotate: bool = False,
#     lp: int = 128,
#     hp: int = 8,
#     sigma: int = 6,
#     ref_mask: np.ndarray = None,
#     xcorr_limit: int = None
# ):
#     import matplotlib.pyplot as plt
#     import numpy as np
#     from fibsem import alignment
#     from fibsem.structures import Point
#     from fibsem.imaging import utils as image_utils

#     # rotate ref
#     if rotate:
#         ref_image = image_utils.rotate_image(ref_image)

#     dx, dy, xcorr = alignment.shift_from_crosscorrelation(
#         ref_image,
#         new_image,
#         lowpass=lp,
#         highpass=hp,
#         sigma=sigma,
#         use_rect_mask=True,
#         ref_mask=ref_mask,
#         xcorr_limit=xcorr_limit
#     )

#     pixelsize = ref_image.metadata.binary_result.pixel_size.x
#     dx_p, dy_p = int(dx / pixelsize), int(dy / pixelsize)

#     print(f"shift_m: {dx}, {dy}")
#     print(f"shift_px: {dx_p}, {dy_p}")

#     shift = np.roll(new_image.data, (-dy_p, -dx_p), axis=(0, 1))

#     mid = Point(shift.shape[1] // 2, shift.shape[0] // 2)

#     if ref_mask is None:
#         ref_mask = np.ones_like(ref_image.data)

#     fig, ax = plt.subplots(1, 4, figsize=(30, 30))
#     ax[0].imshow(ref_image.data * ref_mask , cmap="gray")
#     ax[0].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[0].set_title(f"Reference (rotate={rotate})")
#     ax[1].imshow(new_image.data, cmap="gray")
#     ax[1].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[1].set_title(f"New Image")
#     ax[2].imshow(xcorr, cmap="turbo")
#     ax[2].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[2].plot(mid.x - dx_p, mid.y - dy_p, "m+", ms=50, markeredgewidth=2)
#     ax[2].set_title("XCORR")
#     ax[3].imshow(shift, cmap="gray")
#     ax[3].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2, label="new_position")
#     ax[3].plot(mid.x - dx_p, mid.y - dy_p, "m+", ms=50, markeredgewidth=2, label="old_position")
#     ax[3].set_title("New Image Shifted")
#     ax[3].legend()
#     plt.show()

#     return dx, dy, xcorr


# def plot_crosscorrelation(ref_image, new_image, dx, dy, xcorr):
#     import matplotlib.pyplot as plt
#     from fibsem.structures import Point

#     pixelsize = ref_image.metadata.binary_result.pixel_size.x
#     dx_p, dy_p = int(dx / pixelsize), int(dy / pixelsize)

#     print(f"shift_m: {dx}, {dy}")
#     print(f"shift_px: {dx_p}, {dy_p}")

#     shift = np.roll(new_image.data, (-dy_p, -dx_p), axis=(0, 1))

#     mid = Point(shift.shape[1] // 2, shift.shape[0] // 2)

#     fig, ax = plt.subplots(1, 4, figsize=(30, 30))
#     ax[0].imshow(ref_image.data, cmap="gray")
#     ax[0].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[0].set_title(f"Reference)")
#     ax[1].imshow(new_image.data, cmap="gray")
#     ax[1].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[1].set_title(f"New Image")
#     ax[2].imshow(xcorr, cmap="turbo")
#     ax[2].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[2].plot(mid.x - dx_p, mid.y - dy_p, "m+", ms=50, markeredgewidth=2)
#     ax[2].set_title("XCORR")
#     ax[3].imshow(shift, cmap="gray")
#     ax[3].plot(mid.x, mid.y, color="lime", marker="+", ms=50, markeredgewidth=2)
#     ax[3].plot(mid.x - dx_p, mid.y - dy_p, "m+", ms=50, markeredgewidth=2)
#     ax[3].set_title("New Image Shifted")
#     plt.show()


import logging
from datetime import datetime
import numpy as np
from pathlib import Path

import skimage

from fibsem import acquire, utils
from fibsem.microscope import FibsemMicroscope
from fibsem.structures import (BeamSettings, BeamSystemSettings, BeamType,
                               FibsemRectangle, FibsemStagePosition,
                               ImageSettings, MicroscopeSettings, FibsemImage,
                               MicroscopeState)
from fibsem.detection.detection import NeedleTip, ImageCentre
from fibsem import config as cfg
def auto_focus_beam(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    beam_type: BeamType,
    metric_fn = None, # function to calculate focus metric
    focus_image_settings: ImageSettings = None,
    step_size: float = 0.05e-3,
    num_steps: int = 5,
    kwargs: dict = {},
    verbose: bool = False,
) -> None:

    # TODO: @patrickcleeve2 this could be generalised further if we specify the parameter to sweep through too...
    # e.g. microscope.set("working_distance", value, beam_type) for auto focus
    # e.g. microscope.set("beam_current", value, beam_type) for auto stigmation
    # e.g. microscope.set("beam_current", value, beam_type) for auto beam current
    # might be too much generalisation though...
    # also need to specify a default function

    if metric_fn is None:
        # run the default autofocus routine
        microscope.auto_focus(beam_type=beam_type)
        return
        
    if focus_image_settings is None:
        # use preset settings if not defined
        focus_image_settings = ImageSettings(
            resolution=[768, 512],
            dwell_time=200e-9,
            hfw=100e-6,
            beam_type=beam_type,
            save=True,
            path=settings.image.path,
            autocontrast=True,
            autogamma=False,
            filename=f"{utils.current_timestamp()}_",
            reduced_area=FibsemRectangle(0.3, 0.3, 0.4, 0.4),
        )

    # get current working distance
    current_wd = microscope.get("working_distance", beam_type)

    if verbose:
        logging.info(f"{metric_fn.__name__} based auto-focus routine")
        logging.info(f"doc: {metric_fn.__doc__}")
        logging.info(f"initial working distance: {current_wd:.2e}")

    # define working distance range
    min_wd = current_wd - (num_steps * step_size / 2)
    max_wd = current_wd + (num_steps * step_size / 2)
    wds = np.linspace(min_wd, max_wd, num_steps + 1)

    # loop through working distances and calculate the sharpness (acutance)
    # highest acutance is best focus
    metrics = []
    for i, wd in enumerate(wds):

        logging.info(f"image {i}: {wd:.2e}")
        microscope.set("working_distance", wd, beam_type)

        focus_image_settings.filename = f"{utils.current_timestamp()}_sharpness_{i}"
        img = acquire.new_image(microscope, focus_image_settings)

        # calculate focus metric 
        metric = metric_fn(img, **kwargs)
        metrics.append(metric)

    # select working distance with highest metric
    idx = np.argmax(metrics)

    if verbose:
        pairs = list(zip(wds, metrics))
        logging.info([f"{wd:.2e}: {metric:.4f}" for wd, metric in pairs])
        logging.info(f"{idx}, {wds[idx]:.2e}, {metrics[idx]:.4f}")

    # set working distance
    microscope.set(
        key="working_distance",
        value=wds[idx],
        beam_type=beam_type,
    )


    return

def _sharpness(img:FibsemImage, **kwargs) -> float:
    """Calculate sharpness (accutance) of an image.
    (Acutance: https://en.wikipedia.org/wiki/Acutance)

    Args:
        img (FibsemImage): _description_
        disk_size (int, optional): _description_. Defaults to 5.

    Returns:
        _type_: _description_
    """
    from skimage.filters.rank import gradient
    from skimage.morphology import disk
    disk_size = kwargs.get("disk_size", 5)
    logging.info(f"calculating sharpness (accutance) of image {img}: {disk_size}")
    return np.mean(gradient(skimage.filters.median(np.copy(img.data)), disk(disk_size)))

def _dog(img: FibsemImage, **kwargs) -> float:
    """Calculate difference of gaussian (DoG) of an image.

    Args:
        img (FibsemImage): _description_

    Returns:
        _type_: _description_
    """
    low = kwargs.get("low", 3)
    high = kwargs.get("high", 9)
    from skimage.filters import difference_of_gaussians
    return np.mean(difference_of_gaussians(np.copy(img.data), low, high))

def auto_charge_neutralisation(
    microscope: FibsemMicroscope,
    image_settings: ImageSettings,
    discharge_settings: ImageSettings = None,
    n_iterations: int = 10,
) -> None:

    # take sequence of images quickly,

    # use preset settings if not defined
    if discharge_settings is None:
        discharge_settings = ImageSettings(
            resolution=[768, 512],
            dwell_time=200e-9,
            hfw=image_settings.hfw,
            beam_type=BeamType.ELECTRON,
            save=False,
            autocontrast=False,
            autogamma=False,
            filename=None,
        )

    for i in range(n_iterations):
        acquire.new_image(microscope, discharge_settings)

    # take image
    acquire.new_image(microscope, image_settings)

    logging.info(f"BAM! and the charge is gone!")  # important information


# def auto_needle_calibration(
#     microscope: FibsemMicroscope, settings: MicroscopeSettings, validate: bool = True
# ):

#     if TESCAN:
#         raise NotImplementedError

#     # set coordinate system
#     microscope.connection.specimen.manipulator.set_default_coordinate_system(
#         ManipulatorCoordinateSystem.STAGE
#     )

#     # current working distance
#     wd = microscope.connection.beams.electron_beam.working_distance.value
#     needle_wd_eb = 4.0e-3

#     # focus on the needle
#     microscope.connection.beams.electron_beam.working_distance.value = needle_wd_eb
#     microscope.connection.specimen.stage.link()

#     settings.image.hfw = 2700e-6
#     acquire.take_reference_images(microscope, settings.image)

#     # very low res alignment
#     hfws = [2700e-6, 900e-6, 400e-6, 150e-6]
#     for hfw in hfws:
#         settings.image.hfw = hfw
#         align_needle_to_eucentric_position(microscope, settings, validate=validate)

#     # restore working distance
#     microscope.connection.beams.electron_beam.working_distance.value = wd
#     microscope.connection.specimen.stage.link()

#     logging.info(f"Finished automatic needle calibration.")


# def align_needle_to_eucentric_position(
#     microscope: FibsemMicroscope,
#     settings: MicroscopeSettings,
#     validate: bool = False,
# ) -> None:
#     """Move the needle to the eucentric position, and save the updated position to disk

#     Args:
#         microscope (FibsemMicroscope): OpenFIBSEM microscope instance
#         settings (MicroscopeSettings): microscope settings
#         validate (bool, optional): validate the alignment. Defaults to False.
#     """

#     from fibsem.ui import windows as fibsem_ui_windows
#     from fibsem.detection import detection

#     # take reference images
#     settings.image.save = False
#     settings.image.beam_type = BeamType.ELECTRON

#     det = fibsem_ui_windows.detect_features_v2(
#         microscope=microscope,
#         settings=settings,
#         features=[
#             NeedleTip(),
#             ImageCentre(),
#         ],
#         validate=validate,
#     )
#     detection.move_based_on_detection(
#         microscope, settings, det, beam_type=settings.image.beam_type
#     )

#     # take reference images
#     settings.image.save = False
#     settings.image.beam_type = BeamType.ION

#     image = acquire.new_image(microscope, settings.image)

#     det = fibsem_ui_windows.detect_features_v2(
#         microscope=microscope,
#         settings=settings,
#         features=[
#             NeedleTip(),
#             ImageCentre(),
#         ],
#         validate=validate,
#     )
#     detection.move_based_on_detection(
#         microscope, settings, det, beam_type=settings.image.beam_type, move_x=False
#     )

#     # take image
#     acquire.take_reference_images(microscope, settings.image)


def auto_home_and_link_v2(
    microscope: FibsemMicroscope, state: MicroscopeState = None
) -> None:

    # home the stage and return the linked state
    if state is None:
        state = microscope.get_microscope_state()

    # home the stage
    microscope.home_stage()

    # move to saved linked state
    microscope.set_microscope_state(state)


def _calibrate_manipulator_thermo(microscope:FibsemMicroscope, settings:MicroscopeSettings, parent_ui = None):
    from fibsem.detection import detection
    from fibsem.segmentation.model import load_model
    import matplotlib.pyplot as plt

    from autolamella.workflows.ui import update_detection_ui, ask_user

    if parent_ui:
        ret = ask_user(parent_ui, 
            msg="Please complete the EasyLift alignment procedure in the xT UI until Step 5. Press Continue to proceed.",
            pos="Continue", neg="Cancel")
        if ret is False:
            return
    else:
        input("Please complete the EasyLift alignment procedure in the xT UI until Step 5. Press Enter to proceed.")


    def align_manipulator_to_eucentric(microsscope: FibsemMicroscope, settings:MicroscopeSettings, parent_ui, validate: bool) -> None:
        return NotImplemented

    settings.protocol["options"].get("checkpoint", cfg.__DEFAULT_CHECKPOINT__)
    model = load_model(settings.protocol["options"]["checkpoint"])
    settings.image.autocontrast = True

    hfws = [2000e-6, 900e-6, 400e-6, 150e-6]

    # set working distance
    wd = microscope.get("working_distance", BeamType.ELECTRON)
    microscope.set("working_distance", microscope.system.electron.eucentric_height, BeamType.ELECTRON)

    for hfw in hfws:
        for beam_type in [BeamType.ELECTRON, BeamType.ION]:
            settings.image.hfw = hfw
            settings.image.beam_type = beam_type

            features = [detection.NeedleTip(), detection.ImageCentre()] if np.isclose(microscope.get("scan_rotation", beam_type), 0) else [detection.NeedleTipBottom(), detection.ImageCentre()]
            
            if parent_ui:
                det = update_detection_ui(microscope, settings, features, parent_ui, validate = True, msg = f"Confirm Feature Detection. Press Continue to proceed.")
            else:
                image = acquire.new_image(microscope, settings.image)
                det = detection.detect_features(image, model, features=features, pixelsize=image.metadata.pixel_size.x)
                detection.plot_detection(det)
                ret  = input("continue? (y/n)")
                
                if ret != "y":
                    return

            move_x = bool(beam_type == BeamType.ELECTRON) # ION calibration only in z
            detection.move_based_on_detection(microscope, settings, det, beam_type, move_x=move_x, _move_system="manipulator")

    # restore working distance
    microscope.set("working_distance", wd, BeamType.ELECTRON)

    if parent_ui:
        ask_user(parent_ui, 
            msg="Alignment of EasyLift complete. Please complete the procedure in xT UI. Press Continue to proceed.",
            pos="Continue")
    print(f"The manipulator should now be centred in both beams.")


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemMovementWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(477, 698)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.label_movement_header = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_movement_header.setFont(font)
        self.label_movement_header.setObjectName("label_movement_header")
        self.gridLayout.addWidget(self.label_movement_header, 0, 0, 1, 2)
        self.pushButton_save_position = QtWidgets.QPushButton(Form)
        self.pushButton_save_position.setStyleSheet("color: white; background-color: rgb(85, 170, 0)")
        self.pushButton_save_position.setObjectName("pushButton_save_position")
        self.gridLayout.addWidget(self.pushButton_save_position, 15, 0, 1, 1)
        self.comboBox_positions = QtWidgets.QComboBox(Form)
        self.comboBox_positions.setObjectName("comboBox_positions")
        self.gridLayout.addWidget(self.comboBox_positions, 16, 1, 1, 1)
        self.label_movement_stage_z = QtWidgets.QLabel(Form)
        self.label_movement_stage_z.setObjectName("label_movement_stage_z")
        self.gridLayout.addWidget(self.label_movement_stage_z, 3, 0, 1, 1)
        self.pushButton_refresh_stage_position_data = QtWidgets.QPushButton(Form)
        self.pushButton_refresh_stage_position_data.setObjectName("pushButton_refresh_stage_position_data")
        self.gridLayout.addWidget(self.pushButton_refresh_stage_position_data, 6, 0, 1, 1)
        self.pushButton_move_flat_ion = QtWidgets.QPushButton(Form)
        self.pushButton_move_flat_ion.setObjectName("pushButton_move_flat_ion")
        self.gridLayout.addWidget(self.pushButton_move_flat_ion, 7, 1, 1, 1)
        self.label_positions_header = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_positions_header.setFont(font)
        self.label_positions_header.setObjectName("label_positions_header")
        self.gridLayout.addWidget(self.label_positions_header, 13, 0, 1, 2)
        self.pushButton_go_to = QtWidgets.QPushButton(Form)
        self.pushButton_go_to.setObjectName("pushButton_go_to")
        self.gridLayout.addWidget(self.pushButton_go_to, 18, 0, 1, 1)
        self.label_saved_positions = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_saved_positions.setFont(font)
        self.label_saved_positions.setObjectName("label_saved_positions")
        self.gridLayout.addWidget(self.label_saved_positions, 16, 0, 1, 1)
        self.label_movement_stage_x = QtWidgets.QLabel(Form)
        self.label_movement_stage_x.setObjectName("label_movement_stage_x")
        self.gridLayout.addWidget(self.label_movement_stage_x, 1, 0, 1, 1)
        self.pushButton_move_flat_electron = QtWidgets.QPushButton(Form)
        self.pushButton_move_flat_electron.setObjectName("pushButton_move_flat_electron")
        self.gridLayout.addWidget(self.pushButton_move_flat_electron, 7, 0, 1, 1)
        self.label_movement_instructions = QtWidgets.QLabel(Form)
        self.label_movement_instructions.setMinimumSize(QtCore.QSize(0, 0))
        self.label_movement_instructions.setWordWrap(True)
        self.label_movement_instructions.setObjectName("label_movement_instructions")
        self.gridLayout.addWidget(self.label_movement_instructions, 11, 0, 1, 2)
        self.label_minimap = QtWidgets.QLabel(Form)
        self.label_minimap.setText("")
        self.label_minimap.setObjectName("label_minimap")
        self.gridLayout.addWidget(self.label_minimap, 24, 0, 1, 2)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 23, 0, 1, 1)
        self.lineEdit_position_name = QtWidgets.QLineEdit(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_position_name.sizePolicy().hasHeightForWidth())
        self.lineEdit_position_name.setSizePolicy(sizePolicy)
        self.lineEdit_position_name.setObjectName("lineEdit_position_name")
        self.gridLayout.addWidget(self.lineEdit_position_name, 19, 1, 1, 1)
        self.label_current_position = QtWidgets.QLabel(Form)
        self.label_current_position.setText("")
        self.label_current_position.setObjectName("label_current_position")
        self.gridLayout.addWidget(self.label_current_position, 18, 1, 1, 1)
        self.pushButton_import = QtWidgets.QPushButton(Form)
        self.pushButton_import.setObjectName("pushButton_import")
        self.gridLayout.addWidget(self.pushButton_import, 22, 0, 1, 1)
        self.checkBox_movement_acquire_electron = QtWidgets.QCheckBox(Form)
        self.checkBox_movement_acquire_electron.setChecked(True)
        self.checkBox_movement_acquire_electron.setObjectName("checkBox_movement_acquire_electron")
        self.gridLayout.addWidget(self.checkBox_movement_acquire_electron, 10, 0, 1, 1)
        self.doubleSpinBox_movement_stage_rotation = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_movement_stage_rotation.setMinimum(-360.0)
        self.doubleSpinBox_movement_stage_rotation.setMaximum(360.0)
        self.doubleSpinBox_movement_stage_rotation.setObjectName("doubleSpinBox_movement_stage_rotation")
        self.gridLayout.addWidget(self.doubleSpinBox_movement_stage_rotation, 4, 1, 1, 1)
        self.label_movement_stage_y = QtWidgets.QLabel(Form)
        self.label_movement_stage_y.setObjectName("label_movement_stage_y")
        self.gridLayout.addWidget(self.label_movement_stage_y, 2, 0, 1, 1)
        self.pushButton_load_image_minimap = QtWidgets.QPushButton(Form)
        self.pushButton_load_image_minimap.setObjectName("pushButton_load_image_minimap")
        self.gridLayout.addWidget(self.pushButton_load_image_minimap, 23, 1, 1, 1)
        self.pushButton_export = QtWidgets.QPushButton(Form)
        self.pushButton_export.setObjectName("pushButton_export")
        self.gridLayout.addWidget(self.pushButton_export, 22, 1, 1, 1)
        self.pushButton_remove_position = QtWidgets.QPushButton(Form)
        self.pushButton_remove_position.setStyleSheet("color: white; background-color: rgb(170, 0, 0)")
        self.pushButton_remove_position.setObjectName("pushButton_remove_position")
        self.gridLayout.addWidget(self.pushButton_remove_position, 15, 1, 1, 1)
        self.doubleSpinBox_movement_stage_y = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_movement_stage_y.setDecimals(5)
        self.doubleSpinBox_movement_stage_y.setMinimum(-1e+20)
        self.doubleSpinBox_movement_stage_y.setMaximum(1e+25)
        self.doubleSpinBox_movement_stage_y.setObjectName("doubleSpinBox_movement_stage_y")
        self.gridLayout.addWidget(self.doubleSpinBox_movement_stage_y, 2, 1, 1, 1)
        self.label_positions_header_info = QtWidgets.QLabel(Form)
        self.label_positions_header_info.setObjectName("label_positions_header_info")
        self.gridLayout.addWidget(self.label_positions_header_info, 14, 0, 1, 2)
        self.label_movement_stage_tilt = QtWidgets.QLabel(Form)
        self.label_movement_stage_tilt.setObjectName("label_movement_stage_tilt")
        self.gridLayout.addWidget(self.label_movement_stage_tilt, 5, 0, 1, 1)
        self.pushButton_update_position = QtWidgets.QPushButton(Form)
        self.pushButton_update_position.setObjectName("pushButton_update_position")
        self.gridLayout.addWidget(self.pushButton_update_position, 19, 0, 1, 1)
        self.doubleSpinBox_movement_stage_x = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_movement_stage_x.setDecimals(5)
        self.doubleSpinBox_movement_stage_x.setMinimum(-10000000000.0)
        self.doubleSpinBox_movement_stage_x.setMaximum(1e+17)
        self.doubleSpinBox_movement_stage_x.setObjectName("doubleSpinBox_movement_stage_x")
        self.gridLayout.addWidget(self.doubleSpinBox_movement_stage_x, 1, 1, 1, 1)
        self.label_movement_stage_rotation = QtWidgets.QLabel(Form)
        self.label_movement_stage_rotation.setObjectName("label_movement_stage_rotation")
        self.gridLayout.addWidget(self.label_movement_stage_rotation, 4, 0, 1, 1)
        self.label_movement_options_header = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_movement_options_header.setFont(font)
        self.label_movement_options_header.setObjectName("label_movement_options_header")
        self.gridLayout.addWidget(self.label_movement_options_header, 8, 0, 1, 2)
        self.checkBox_movement_acquire_ion = QtWidgets.QCheckBox(Form)
        self.checkBox_movement_acquire_ion.setChecked(True)
        self.checkBox_movement_acquire_ion.setObjectName("checkBox_movement_acquire_ion")
        self.gridLayout.addWidget(self.checkBox_movement_acquire_ion, 10, 1, 1, 1)
        self.pushButton_move = QtWidgets.QPushButton(Form)
        self.pushButton_move.setObjectName("pushButton_move")
        self.gridLayout.addWidget(self.pushButton_move, 6, 1, 1, 1)
        self.doubleSpinBox_movement_stage_z = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_movement_stage_z.setDecimals(5)
        self.doubleSpinBox_movement_stage_z.setMinimum(-1e+17)
        self.doubleSpinBox_movement_stage_z.setMaximum(1e+23)
        self.doubleSpinBox_movement_stage_z.setObjectName("doubleSpinBox_movement_stage_z")
        self.gridLayout.addWidget(self.doubleSpinBox_movement_stage_z, 3, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 25, 0, 1, 2)
        self.doubleSpinBox_movement_stage_tilt = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_movement_stage_tilt.setObjectName("doubleSpinBox_movement_stage_tilt")
        self.gridLayout.addWidget(self.doubleSpinBox_movement_stage_tilt, 5, 1, 1, 1)
        self.label_movement_images = QtWidgets.QLabel(Form)
        self.label_movement_images.setObjectName("label_movement_images")
        self.gridLayout.addWidget(self.label_movement_images, 9, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_movement_header.setText(_translate("Form", "Stage Movement"))
        self.pushButton_save_position.setText(_translate("Form", "Add Position"))
        self.label_movement_stage_z.setText(_translate("Form", "Z Coordinate (mm)"))
        self.pushButton_refresh_stage_position_data.setText(_translate("Form", "Refresh Stage Position Data"))
        self.pushButton_move_flat_ion.setText(_translate("Form", "Move Flat to ION Beam"))
        self.label_positions_header.setText(_translate("Form", "Stage Positions"))
        self.pushButton_go_to.setText(_translate("Form", "Go To Position"))
        self.label_saved_positions.setText(_translate("Form", "Saved Positions"))
        self.label_movement_stage_x.setText(_translate("Form", "X Coordinate (mm)"))
        self.pushButton_move_flat_electron.setText(_translate("Form", "Move Flat to ELECTRON Beam"))
        self.label_movement_instructions.setText(_translate("Form", "Double click to move. Press continue when complete."))
        self.label_4.setText(_translate("Form", "Please load grid image to display minimap "))
        self.pushButton_import.setText(_translate("Form", "Import Positions"))
        self.checkBox_movement_acquire_electron.setText(_translate("Form", "Electron Beam"))
        self.label_movement_stage_y.setText(_translate("Form", "Y Coordinate (mm)"))
        self.pushButton_load_image_minimap.setText(_translate("Form", "Load Minimap Image"))
        self.pushButton_export.setText(_translate("Form", "Export Positions"))
        self.pushButton_remove_position.setText(_translate("Form", "Remove Position"))
        self.label_positions_header_info.setText(_translate("Form", "All positions in mm and degrees"))
        self.label_movement_stage_tilt.setText(_translate("Form", "Tilt (deg)"))
        self.pushButton_update_position.setText(_translate("Form", "Update Position"))
        self.label_movement_stage_rotation.setText(_translate("Form", "Rotation (deg)"))
        self.label_movement_options_header.setText(_translate("Form", "Movement Options"))
        self.checkBox_movement_acquire_ion.setText(_translate("Form", "Ion Beam"))
        self.pushButton_move.setText(_translate("Form", "Move to Position"))
        self.label_movement_images.setText(_translate("Form", "Acquire images after moving:"))



import glob
import json
import os
import torch

import argparse


def export_model_checkpoint(
    path: str,
    checkpoint_path: str = None,
    checkpoint_name: str = "model_checkpoint.pth",
) -> None:
    """Save NNUNet model checkpoint as a single .pth file
    args:
        path: path to the nnunet model directory

    """
    # nnunet model directory structure for ensemble:
    # model
    #  dataset.json
    #  plans.json
    #  fold_n:
    #   checkpoint_best.pth
    #   checkpoint_final.pth

    # we want to convert it to a single .pth file with the following structure:
    # model_checkpoint.pth
    # dataset: dataset.json
    # plans: plans.json
    # fold_n:
    #  best:  checkpoint_best.pth
    #  final: checkpoint_final.pth

    # this makes it more portable and easier to load

    def load_json(path: str):
        with open(path, "r") as f:
            return json.load(f)

    # confirm that the path is a nnunet model directory
    if not os.path.isdir(path):
        raise ValueError(f"{path} is not a directory")
    if not os.path.exists(os.path.join(path, "dataset.json")):
        raise ValueError(f"{path} does not contain a dataset.json file")
    if not os.path.exists(os.path.join(path, "plans.json")):
        raise ValueError(f"{path} does not contain a plans.json file")

    print(f"Exporting model checkpoint from {path}...")

    MODEL_CHECKPOINT = {}

    # paths
    DATASET_JSON_PATH = os.path.join(path, "dataset.json")
    PLAN_JSON_PATH = os.path.join(path, "plans.json")

    # load the dataset and plans
    print("Loading dataset and plans configurations...")
    MODEL_CHECKPOINT["dataset"] = load_json(DATASET_JSON_PATH)
    MODEL_CHECKPOINT["plans"] = load_json(PLAN_JSON_PATH)

    # load the folds
    MODEL_CHECKPOINT["folds"] = {}

    # get all the fold directories,
    FOLD_DIRS = sorted(glob.glob(os.path.join(path, "fold_*")))
    print(f"Found {len(FOLD_DIRS)} folds...")
    for fold_dir in FOLD_DIRS:
        fold_name = os.path.basename(fold_dir)
        print(f"Processing fold {fold_name}...")

        # load the best/ final checkpoint
        BEST_CHECKPOINT_PATH = os.path.join(fold_dir, "checkpoint_best.pth")
        FINAL_CHECKPOINT_PATH = os.path.join(fold_dir, "checkpoint_final.pth")

        MODEL_CHECKPOINT["folds"][fold_name] = {
            "best": torch.load(BEST_CHECKPOINT_PATH, map_location=torch.device("cpu")),
            "final": torch.load(
                FINAL_CHECKPOINT_PATH, map_location=torch.device("cpu")
            ),
        }

    # save as single torch checkpoint
    if checkpoint_path is None:
        checkpoint_path = os.path.join(path, checkpoint_name)
    torch.save(MODEL_CHECKPOINT, checkpoint_path)
    print(f"Saved model checkpoint to {checkpoint_path}")


def main():
    parser = argparse.ArgumentParser(
        description="Export nnunet model checkpoint as a single .pth file"
    )
    parser.add_argument("--path", type=str, help="path to nnunet model directory")
    parser.add_argument(
        "--checkpoint_path",
        type=str,
        help="path to save the checkpoint",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--checkpoint_name",
        type=str,
        help="name of the checkpoint",
        required=False,
        default="model_checkpoint.pth",
    )

    args = parser.parse_args()

    export_model_checkpoint(
        path=args.path,
        checkpoint_path=args.checkpoint_path,
        checkpoint_name=args.checkpoint_name,
    )


if __name__ == "__main__":
    main()



from fibsem.microscope import _THERMO_API_AVAILABLE, _TESCAN_API_AVAILABLE



def main():

    # OpenFIBSEM API
    print(f"\n\nOpenFIBSEM API:\n")
    try:
        import fibsem
        FIBSEM_AVAILABLE = True
    except ImportError:
        FIBSEM_AVAILABLE = False
    if FIBSEM_AVAILABLE:
        print(f"OpenFIBSEM v{fibsem.__version__}")
        print(f"Installed at: {fibsem.__path__}")
    
    print(f"-" * 80)
    
    print(f"Applications:\n")
    try: 
        import autolamella 
        AUTOLAMELLA_AVAILABLE = True
    except ImportError: 
        AUTOLAMELLA_AVAILABLE = False
    if AUTOLAMELLA_AVAILABLE:
        print(f"AutoLamella v{autolamella.__version__}")
        print(f"Installed at: {autolamella.__path__}")   
    
    try: 
        import salami
        SALAMI_AVAILABLE = True
    except ImportError:
        SALAMI_AVAILABLE = False
    if SALAMI_AVAILABLE:
        print(f"SALAMI v{salami.__version__}")
        print(f"Installed at: {salami.__path__}")
    print(f"-" * 80)
    
    # Hardware APIs
    print(f"Hardware APIs:\n")
    
    # Thermo Fisher API
    print(f"ThermoFisher API {'Available' if _THERMO_API_AVAILABLE else 'Not Available'}")
    if _THERMO_API_AVAILABLE:
        from fibsem.microscope import version as autoscript_version
        print(f"AutoScript v{autoscript_version}")
    print(f"-" * 80)

    # Tescan API
    print(f"Tescan API {'Available' if _TESCAN_API_AVAILABLE else 'Not Available'}")
    if _TESCAN_API_AVAILABLE:
        from fibsem.microscope import tescanautomation
        print(f"TescanAutomation v{tescanautomation.__version__}")

if __name__ == "__main__":
    main()


# 
from langchain.document_loaders import UnstructuredPDFLoader
from langchain.indexes import VectorstoreIndexCreator
from langchain.llms import OpenAI
import os

# import OpenAI API key from secret.txt
with open("secret.txt", "r") as f:
    os.environ["OPENAI_API_KEY"] = f.read()

def create_index(filenames: list[str]):
    # TODO: make a list of files to index
    llm = OpenAI(openai_api_key="OPENAI_API_KEY")
    loaders = [UnstructuredPDFLoader(fname) for fname in filenames]
    index = VectorstoreIndexCreator().from_loaders(loaders)

    return index 


# TODO: select files
# TODO: tool use

# persist
# https://python.langchain.com/en/latest/modules/indexes/vectorstores/examples/chroma.html

# enter an endless loop with user input
def main():
    print("Welcome to FIBSEM Chat! Enter your question, or type 'quit' to exit.")

    # load user manual, index
    filenames = ["documents/OnlineManualHeliosHydra.pdf"]
    # filenames = ["documents/autoliftout.pdf", "documents/supplementary.pdf"]

    print(f"Creating index from {filenames}.")
    index =  create_index(filenames=filenames)
    print(f"Index created from {filenames}.\n")

    while True:

        q = input("Enter a question: ")

        if q in ["quit", "exit"]:
            print("\nGoodbye!")
            break
        
        print("Thinking...")

        # run query, show result
        ret = index.query(q)   

        print("\n\nResponse: " + ret)
        print("-"*50)


if __name__ == "__main__":
    main()



import pytest
import numpy as np

from fibsem import movement

def test_angle_difference():
    assert np.isclose(movement.angle_difference(np.deg2rad(0), np.deg2rad(0)), 0)
    assert np.isclose(movement.angle_difference(np.deg2rad(0), np.deg2rad(360)), 0)
    assert np.isclose(movement.angle_difference(np.deg2rad(0), np.deg2rad(180)), np.pi)
    assert np.isclose(movement.angle_difference(np.deg2rad(2), np.deg2rad(358)), np.deg2rad(4))
    assert np.isclose(movement.angle_difference(np.deg2rad(-45), np.deg2rad(45)), np.pi / 2)
    assert np.isclose(movement.angle_difference(np.deg2rad(-360), np.deg2rad(360)), 0)
    assert np.isclose(movement.angle_difference(-4*np.pi, 4*np.pi), 0)

def test_rotation_angle_is_larger():
    assert movement.rotation_angle_is_larger(np.deg2rad(0), np.deg2rad(0)) == False
    assert movement.rotation_angle_is_larger(np.deg2rad(0), np.deg2rad(360)) == False
    assert movement.rotation_angle_is_larger(np.deg2rad(-45), np.deg2rad(45)) == False
    assert movement.rotation_angle_is_larger(np.deg2rad(0), np.deg2rad(180)) == True
    assert movement.rotation_angle_is_larger(np.deg2rad(-90), np.deg2rad(90)) == True
    assert movement.rotation_angle_is_larger(np.deg2rad(0), np.deg2rad(720)) == False


def test_rotation_angle_is_smaller():
    assert movement.rotation_angle_is_smaller(np.deg2rad(0), np.deg2rad(0), 5) == True
    assert movement.rotation_angle_is_smaller(np.deg2rad(0), np.deg2rad(360), 5) == True
    assert movement.rotation_angle_is_smaller(np.deg2rad(0), np.deg2rad(180)) == False
    assert movement.rotation_angle_is_smaller(np.deg2rad(-90), np.deg2rad(90), 180) == False
    assert movement.rotation_angle_is_smaller(np.deg2rad(0), np.deg2rad(-360), 1) == True





from fibsem import utils, acquire

import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('TkAgg', force=True) # Activate 'agg' backend for off-screen plotting.


def main():

    # connect to microscope
    microscope, settings = utils.setup_session(manufacturer="Demo", ip_address="localhost")

    # take image with both beams
    eb_image, ib_image = acquire.take_reference_images(microscope, settings.image)

    # show images
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    ax[0].imshow(eb_image.data, cmap="gray")
    ax[0].set_title("Electron Beam Image")
    ax[0].axis("off")
    ax[1].imshow(ib_image.data, cmap="gray")
    ax[1].set_title("Ion Beam Image")
    ax[1].axis("off")
    plt.show()


if __name__ == "__main__":
    main()



import logging
import napari
import napari.utils.notifications
import numpy as np
from PyQt5 import QtWidgets
from fibsem import config as cfg
from fibsem import constants, conversions, utils
from fibsem.microscope import FibsemMicroscope, TescanMicroscope, ThermoMicroscope, DemoMicroscope
from fibsem.structures import (MicroscopeSettings)
from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
from fibsem.ui.qtdesigner_files import FibsemGISWidget
from fibsem.gis import deposit_platinum


class FibsemGISWidget(FibsemGISWidget.Ui_Form, QtWidgets.QWidget):
    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        viewer: napari.Viewer = None,
        image_widget: FibsemImageSettingsWidget = None, 
        parent=None,
    ):
        super(FibsemGISWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.microscope = microscope
        self.settings = settings
        self.viewer = viewer
        self.image_widget = image_widget

        self.gas_protocol = {}

        self.gis_and_mc_setup()

        self.initial_gas_setup()

        self.setup_connections()

        self.update_ui()


    def gis_and_mc_setup(self):

        if isinstance(self.microscope, ThermoMicroscope):
            self.gis_lines = self.microscope.multichem_available_lines()
        else:
            self.gis_lines = self.microscope.GIS_available_lines()
       

        self.gas_combobox.addItems(self.gis_lines)
        self.gis_current_line = self.gis_lines[0]

        self.temp_label.setText(f"Temp: Need Warm Up")
        self.warm_button.setEnabled(True)

        self.gas_combobox.setCurrentText(self.gis_current_line)

        if isinstance(self.microscope, ThermoMicroscope):
            self.gis_available_positions = self.microscope.multichem_available_positions()
        else:
            self.gis_available_positions = self.microscope.GIS_available_positions()
        
        self.position_combobox.addItems(self.gis_available_positions)

        protocol = utils.load_yaml(cfg.PROTOCOL_PATH)
        self.protocol = protocol

        if isinstance(self.microscope, TescanMicroscope):
            self.tescan_setup()
        if isinstance(self.microscope, (ThermoMicroscope, DemoMicroscope)):
            self.thermo_setup()


    def initial_gas_setup(self):

        self.gas_protocol = {
            "application_file": self.protocol["milling"]["application_file"],
            "gas": self.gis_current_line,
            "position": "Retract",
            "hfw":self.image_widget.image_settings.hfw,
            "length": 7.0e-6,
            "spot_size": 5.0e-8,
            "beam_current": 5.0e-10,
            "dwell_time": 1.0e-6,
            "beam_type": "ELECTRON",
            "time": 1.5,
            "blank_beam": False,
        }

        self.timeDuration_spinbox.setValue(self.gas_protocol["time"])

    def tescan_setup(self):
        
        self.app_file_combobox.hide()
        self.app_file_label.hide()
        self.blankBeamcheckbox.hide()
        self.timeDuration_spinbox.setValue(1)
        self.update_ui()

    def thermo_setup(self):

        self.gas_combobox.clear()
        self.gas_combobox.addItems(self.gis_lines)
        self.gas_combobox.setCurrentText(self.gis_current_line)

        self.blankBeamcheckbox.setChecked(False)


        app_files =  self.microscope.get_available_values(key="application_file")

        self.app_file_combobox.addItems(app_files)

        if self.protocol['milling']['application_file'] in app_files:
            self.app_file_combobox.setCurrentText(self.protocol['milling']['application_file'])
        else:
            self.app_file_combobox.setCurrentText(app_files[0])
            self.gas_protocol['application_file'] = app_files[0]
            logging.info(f"Application file {self.protocol['milling']['application_file']} not found in microscope")
            logging.info(f'Application file set to {app_files[0]}')

        self.update_ui()


    def update_gas_protocol(self):

        self.gas_protocol["application_file"] = self.app_file_combobox.currentText()
        self.gas_protocol["beam_type"] = self.beamtype_combobox.currentText()
        self.gas_protocol["hfw"] = self.hfw_spinbox.value()*constants.MICRON_TO_METRE
        self.gas_protocol["time"] = self.timeDuration_spinbox.value()
        self.gas_protocol["blank_beam"] = self.blankBeamcheckbox.isChecked()
        self.gas_protocol["position"] = self.position_combobox.currentText()
        self.gas_protocol["gas"] = self.gas_combobox.currentText()

    def setup_connections(self):
        self.gas_combobox.currentIndexChanged.connect(self.update_ui)
        self.move_GIS_button.clicked.connect(self.move_gis)
        self.warm_button.clicked.connect(self.warm_up_gis)
        self.run_button.clicked.connect(self.run_gis)
        self.beamtype_combobox.setCurrentText("ION")
        self.pushButton_settings_update.clicked.connect(self.update_gas_protocol)

    

    def warm_up_gis(self):
        
        line_name = self.gis_current_line

        self.microscope.multichem_heat_up(line_name) if isinstance(self.microscope, ThermoMicroscope) else self.microscope.GIS_heat_up(line_name)
        self.temp_label.setText("Temp: Ready")


    def move_gis(self):

        position = self.position_combobox.currentText()
        if isinstance(self.microscope, ThermoMicroscope):
            self.microscope.multichem_move_to(position)
        else:
            self.microscope.GIS_move_to(self.gis_current_line,position)
        
        self.gas_protocol["position"] = position
        self.update_ui()

    def update_ui(self):

        line_name = self.gas_combobox.currentText()
        if line_name == "":
            return

        self.gis_current_line = line_name
       
            
        if isinstance(self.microscope, ThermoMicroscope):
            current_position =self.microscope.multichem_position()
            temp_ready = "Ready" if self.microscope.multichem_temp_ready(self.gis_current_line) else "Need Warm Up"
        else:
            temp_ready = "Ready" if self.microscope.GIS_temp_ready(self.gis_current_line) else "Need Warm Up"
            current_position = self.microscope.GIS_position(self.gis_current_line)

        self.temp_label.setText(f"Temp: {temp_ready}")

        self.position_combobox.setCurrentText(current_position)
        self.current_position_label.setText(f"Current Position: {current_position}")
        self.hfw_spinbox.setValue(self.image_widget.image_settings.hfw*constants.METRE_TO_MICRON)



    def run_gis(self):

        self.gas_protocol["gas"] = self.gis_current_line
        deposit_platinum(self.microscope, self.gas_protocol)
        
        


def main(): 

    viewer = napari.Viewer(ndisplay=2)
    movement_widget = FibsemGISWidget()
    viewer.window.add_dock_widget(
        movement_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()


import logging

import numpy as np

from scipy import fftpack

from fibsem import acquire, calibration, utils, validation
from fibsem.imaging import masks
from fibsem.imaging import utils as image_utils
from fibsem.structures import (
    BeamType,
    ImageSettings,
    MicroscopeSettings,
    ReferenceImages,
    FibsemImage,
    FibsemRectangle,
)
from fibsem.microscope import FibsemMicroscope
from typing import Union, Optional


def auto_eucentric_correction(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    image_settings: ImageSettings,
    tilt_degrees: int = 25,
    xcorr_limit: int = 250,
) -> None:

    raise NotImplementedError

    image_settings.save = False
    image_settings.beam_type = BeamType.ELECTRON
    calibration.auto_charge_neutralisation(
        microscope.connection, image_settings
    )  # TODO: need to change this function

    for hfw in [400e-6, 150e-6, 80e-6, 80e-6]:
        image_settings.hfw = hfw

        correct_stage_eucentric_alignment(
            microscope,
            settings,
            image_settings,
            tilt_degrees=tilt_degrees,
            xcorr_limit=xcorr_limit,
        )


def beam_shift_alignment(
    microscope: FibsemMicroscope,
    image_settings: ImageSettings,
    ref_image: FibsemImage,
    reduced_area: Optional[FibsemRectangle] = None,
    alignment_current: Optional[float] = None,
):
    """Aligns the images by adjusting the beam shift instead of moving the stage.

    This method uses cross-correlation between the reference image and a new image to calculate the
    optimal beam shift for alignment. This approach offers increased precision, but a lower range
    compared to stage movement.

    NOTE: Only shift the ion beam, never the electron beam.

    Args:
        microscope (FibsemMicroscope): An OpenFIBSEM microscope client.
        image_settings (acquire.ImageSettings): Settings for taking the image.
        ref_image (FibsemImage): The reference image to align to.
        reduced_area (FibseRectangle): The reduced area to image with.

    Raises:
        ValueError: If `image_settings.beam_type` is not set to `BeamType.ION`.

    """
    if alignment_current is not None:
        initial_current = microscope.get("current", image_settings.beam_type)
        microscope.set("current", alignment_current, image_settings.beam_type)

    # TODO: we should just use the same settings as the reference image, no need to pass in image_settings?
    
    import time
    time.sleep(3) # threading is too fast?
    ref_image_settings = ImageSettings.fromFibsemImage(ref_image)
    ref_image_settings.beam_type = BeamType.ION
    ref_image_settings.reduced_area = reduced_area
    ref_image_settings.autocontrast = False
    ref_image_settings.save = True
    ref_image_settings.filename = image_settings.filename
    new_image = acquire.new_image(
        microscope, settings=ref_image_settings
    )
    dx, dy, _ = shift_from_crosscorrelation(
        ref_image, new_image, lowpass=50, highpass=4, sigma=5, use_rect_mask=True
    )
    image_settings.autocontrast = True
    

    # adjust beamshift (reverse direction for x)
    microscope.beam_shift(-dx, dy, image_settings.beam_type)

    # reset beam current
    if alignment_current is not None:
        microscope.set("current", initial_current, image_settings.beam_type)


def beam_shift_alignment_v2(
    microscope: FibsemMicroscope,
    ref_image: FibsemImage,
    alignment_current: Optional[float] = None,
):
    """Aligns the images by adjusting the beam shift instead of moving the stage.

    This method uses cross-correlation between the reference image and a new image to calculate the
    optimal beam shift for alignment. This approach offers increased precision, but a lower range
    compared to stage movement.

    NOTE: Only shift the ion beam, never the electron beam.

    Args:
        microscope (FibsemMicroscope): An OpenFIBSEM microscope client.
        ref_image (FibsemImage): The reference image to align to.

    Raises:
        ValueError: If `image_settings.beam_type` is not set to `BeamType.ION`.

    """

    import time
    time.sleep(3) # threading is too fast?
    image_settings = ImageSettings.fromFibsemImage(ref_image)
    image_settings.autocontrast = False
    image_settings.save = True
    image_settings.filename = f"beam_shift_alignment_{utils.current_timestamp_v2()}"

    # set alignment current
    if alignment_current is not None:
        initial_current = microscope.get("current", image_settings.beam_type)
        microscope.set("current", alignment_current, image_settings.beam_type)

    new_image = acquire.new_image(
        microscope, settings=image_settings
    )
    dx, dy, _ = shift_from_crosscorrelation(
        ref_image, new_image, lowpass=50, highpass=4, sigma=5, use_rect_mask=True
    )    

    # adjust beamshift (reverse direction)
    microscope.beam_shift(-dx, -dy, image_settings.beam_type)

    # reset beam current
    if alignment_current is not None:
        microscope.set("current", initial_current, image_settings.beam_type)
    

def correct_stage_drift(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    reference_images: ReferenceImages,
    alignment: tuple[BeamType, BeamType] = (BeamType.ELECTRON, BeamType.ELECTRON),
    rotate: bool = False,
    ref_mask_rad: int = 512,
    xcorr_limit: Union[tuple[int, int], None] = None,
    constrain_vertical: bool = False,
    use_beam_shift: bool = False,
) -> bool:
    """Corrects the stage drift by aligning low- and high-resolution reference images
    using cross-correlation.

    Args:
        microscope (FibsemMicroscope): The microscope used for image acquisition.
        settings (MicroscopeSettings): The settings used for image acquisition.
        reference_images (ReferenceImages): A container of low- and high-resolution
            reference images.
        alignment (tuple[BeamType, BeamType], optional): A tuple of two `BeamType`
            objects, specifying the beam types used for the alignment of low- and
            high-resolution images, respectively. Defaults to (BeamType.ELECTRON,
            BeamType.ELECTRON).
        rotate (bool, optional): Whether to rotate the reference images before
            alignment. Defaults to False.
        ref_mask_rad (int, optional): The radius of the circular mask used for reference
        xcorr_limit (tuple[int, int] | None, optional): A tuple of two integers that
            represent the minimum and maximum cross-correlation values allowed for the
            alignment. If not specified, the values are set to (None, None), which means
            there are no limits. Defaults to None.
        constrain_vertical (bool, optional): Whether to constrain the alignment to the
            vertical axis. Defaults to False.

    Returns:
        bool: True if the stage drift correction was successful, False otherwise.
    """

    # set reference images
    if alignment[0] is BeamType.ELECTRON:
        ref_lowres, ref_highres = (
            reference_images.low_res_eb,
            reference_images.high_res_eb,
        )
    if alignment[0] is BeamType.ION:
        ref_lowres, ref_highres = (
            reference_images.low_res_ib,
            reference_images.high_res_ib,
        )

    if xcorr_limit is None:
        xcorr_limit = (None, None)

    # rotate reference
    if rotate:
        ref_lowres = image_utils.rotate_image(ref_lowres)
        ref_highres = image_utils.rotate_image(ref_highres)

    # align lowres, then highres
    for i, ref_image in enumerate([ref_lowres, ref_highres]):

        ref_mask = masks.create_circle_mask(ref_image.data.shape, ref_mask_rad)

        # take new images
        # set new image settings (same as reference)
        settings.image = ImageSettings.fromFibsemImage(ref_image)
        settings.image.beam_type = alignment[1]
        new_image = acquire.new_image(microscope, settings.image)

        # crosscorrelation alignment
        ret = align_using_reference_images(
            microscope,
            settings,
            ref_image,
            new_image,
            ref_mask=ref_mask,
            xcorr_limit=xcorr_limit[i],
            constrain_vertical=constrain_vertical,
            use_beam_shift=use_beam_shift,
        )

        if ret is False:
            break  # cross correlation has failed...

    return ret


def align_using_reference_images(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    ref_image: FibsemImage,
    new_image: FibsemImage,
    ref_mask: np.ndarray = None,
    xcorr_limit: int = None,
    constrain_vertical: bool = False,
    use_beam_shift: bool = False,
) -> bool:
    """
    Uses cross-correlation to align a new image to a reference image.

    Args:
        microscope: A FibsemMicroscope instance representing the microscope being used.
        settings: A MicroscopeSettings instance representing the settings for the imaging session.
        ref_image: A FibsemImage instance representing the reference image to which the new image will be aligned.
        new_image: A FibsemImage instance representing the new image that will be aligned to the reference image.
        ref_mask: A numpy array representing a mask to apply to the reference image during alignment. Default is None.
        xcorr_limit: An integer representing the limit for the cross-correlation coefficient. If the coefficient is below
            this limit, alignment will fail. Default is None.
        constrain_vertical: A boolean indicating whether to constrain movement to the vertical axis. If True, movement
            will be restricted to the vertical axis, which is useful for eucentric movement. If False, movement will be
            allowed on both the X and Y axes. Default is False.

    Returns:
        A boolean indicating whether the alignment was successful. True if the alignment was successful, False otherwise.
    """
    # get beam type
    ref_beam_type = BeamType[ref_image.metadata.image_settings.beam_type.name.upper()]
    new_beam_type = BeamType[new_image.metadata.image_settings.beam_type.name.upper()]

    logging.info(
        f"aligning {ref_beam_type.name} reference image to {new_beam_type.name}."
    )
    sigma = 6
    hp_px = 8
    lp_px = 128  # MAGIC_NUMBER

    dx, dy, xcorr = shift_from_crosscorrelation(
        ref_image,
        new_image,
        lowpass=lp_px,
        highpass=hp_px,
        sigma=sigma,
        use_rect_mask=True,
        ref_mask=ref_mask,
        xcorr_limit=xcorr_limit,
    )

    shift_within_tolerance = (
        validation.check_shift_within_tolerance(  # TODO: Abstract validation.py
            dx=dx, dy=dy, ref_image=ref_image, limit=0.5
        )
    )

    if shift_within_tolerance:

        # vertical constraint = eucentric movement
        if constrain_vertical:
            microscope.vertical_move( dx=0, dy=-dy
            )  # FLAG_TEST
        else:
            if use_beam_shift:
                # move the beam shift
                microscope.beam_shift(dx=-dx, dy=-dy, beam_type=new_beam_type)
            else:
                # move the stage
                microscope.stable_move(
                    dx=dx,
                    dy=-dy,
                    beam_type=new_beam_type,
                )

    return shift_within_tolerance


def shift_from_crosscorrelation(
    ref_image: FibsemImage,
    new_image: FibsemImage,
    lowpass: int = 128,
    highpass: int = 6,
    sigma: int = 6,
    use_rect_mask: bool = False,
    ref_mask: np.ndarray = None,
    xcorr_limit: int = None,
) -> tuple[float, float, np.ndarray]:
    """Calculates the shift between two images by cross-correlating them and finding the position of maximum correlation.

    Args:
        ref_image (FibsemImage): The reference image.
        new_image (FibsemImage): The new image to align to the reference.
        lowpass (int, optional): The low-pass filter frequency (in pixels) for the bandpass filter used to
            enhance the correlation signal. Defaults to 128.
        highpass (int, optional): The high-pass filter frequency (in pixels) for the bandpass filter used to
            enhance the correlation signal. Defaults to 6.
        sigma (int, optional): The standard deviation (in pixels) of the Gaussian filter used to create the bandpass
            mask. Defaults to 6.
        use_rect_mask (bool, optional): Whether to use a rectangular mask for the correlation. If True, the correlation
            is performed only inside a rectangle that covers most of the image, to reduce the effect of noise at the
            edges. Defaults to False.
        ref_mask (np.ndarray, optional): A mask to apply to the reference image before correlation. If not None,
            it should be a binary array with the same shape as the images. Pixels with value 0 will be ignored in the
            correlation. Defaults to None.
        xcorr_limit (int, optional): If not None, the correlation map will be circularly masked to a square
            with sides of length 2 * xcorr_limit + 1, centred on the maximum correlation peak. This can be used to
            limit the search range and improve the accuracy of the shift. Defaults to None.

    Returns:
        A tuple (x_shift, y_shift, xcorr), where x_shift and y_shift are the shifts along x and y (in meters),
        and xcorr is the cross-correlation map between the images.
    """
    # get pixel_size
    pixelsize_x = new_image.metadata.pixel_size.x
    pixelsize_y = new_image.metadata.pixel_size.y

    # normalise both images
    ref_data_norm = image_utils.normalise_image(ref_image)
    new_data_norm = image_utils.normalise_image(new_image)

    # cross-correlate normalised images
    if use_rect_mask:
        rect_mask = masks._mask_rectangular(new_data_norm.shape)
        ref_data_norm = rect_mask * ref_data_norm
        new_data_norm = rect_mask * new_data_norm

    if ref_mask is not None:
        ref_data_norm = ref_mask * ref_data_norm  # mask the reference

    # bandpass mask
    bandpass = masks.create_bandpass_mask(
        shape=ref_data_norm.shape, lp=lowpass, hp=highpass, sigma=sigma
    )

    # crosscorrelation
    xcorr = crosscorrelation_v2(ref_data_norm, new_data_norm, bandpass=bandpass)

    # limit xcorr range
    if xcorr_limit:
        xcorr = masks.apply_circular_mask(xcorr, xcorr_limit)

    # calculate maximum crosscorrelation
    maxX, maxY = np.unravel_index(np.argmax(xcorr), xcorr.shape)  # TODO: backwards
    cen = np.asarray(xcorr.shape) / 2
    err = np.array(cen - [maxX, maxY], int)

    # calculate shift in metres
    dx = err[1] * pixelsize_x
    dy = err[0] * pixelsize_y  # this could be the issue?

    msgd = {"msg": "cross-correlation", "pixelsize": (pixelsize_x, pixelsize_y), 
        "max": (maxX, maxY), "centre": cen, "shift": (err[1], err[0]), "shift_meters": (dx, dy)}
    logging.debug(msgd)

    # logging.debug(f"cross-correlation:")
    # logging.debug(f"pixelsize: x: {pixelsize_x:.2e}, y: {pixelsize_y:.2e}")
    # logging.debug(f"maxX: {maxX}, {maxY}, centre: {cen}")
    # logging.debug(f"x: {err[1]}px, y: {err[0]}px")
    # logging.debug(f"x: {dx:.2e}m, y: {dy:.2e} meters")

    # save data
    _save_alignment_data(
        ref_image=ref_image,
        new_image=new_image,
        bandpass=bandpass,
        xcorr=xcorr,
        use_rect_mask=use_rect_mask,
        ref_mask=ref_mask,
        xcorr_limit=xcorr_limit,
        lowpass=lowpass,
        highpass=highpass,
        sigma=sigma,
        dx=dx,
        dy=dy,
        pixelsize_x=pixelsize_x,
        pixelsize_y=pixelsize_y,
    )

    # metres
    return dx, dy, xcorr


def crosscorrelation_v2(
    img1: np.ndarray, img2: np.ndarray, bandpass: np.ndarray = None
) -> np.ndarray:
    """
    Cross-correlate two images using Fourier convolution matching.

    Args:
        img1 (np.ndarray): The reference image.
        img2 (np.ndarray): The new image to be cross-correlated with the reference.
        bandpass (np.ndarray, optional): A bandpass mask to apply to both images before cross-correlation. Defaults to None.

    Returns:
        np.ndarray: The cross-correlation map between the two images.
    """
    if img1.shape != img2.shape:
        err = (
            f"Image 1 {img1.shape} and Image 2 {img2.shape} need to have the same shape"
        )
        logging.error(err)
        raise ValueError(err)

    if bandpass is None:
        bandpass = np.ones_like(img1)

    n_pixels = img1.shape[0] * img1.shape[1]

    img1ft = np.fft.ifftshift(bandpass * np.fft.fftshift(np.fft.fft2(img1)))
    tmp = img1ft * np.conj(img1ft)
    img1ft = n_pixels * img1ft / np.sqrt(tmp.sum())

    img2ft = np.fft.ifftshift(bandpass * np.fft.fftshift(np.fft.fft2(img2)))
    img2ft[0, 0] = 0
    tmp = img2ft * np.conj(img2ft)

    img2ft = n_pixels * img2ft / np.sqrt(tmp.sum())

    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots(1, 2, figsize=(15, 15))
    # ax[0].imshow(np.fft.ifft2(img1ft).real)
    # ax[1].imshow(np.fft.ifft2(img2ft).real)
    # plt.show()

    # plt.title("Power Spectra")
    # plt.imshow(np.log(np.abs(np.fft.fftshift(np.fft.fft2(img1)))))
    # plt.show()

    xcorr = np.real(np.fft.fftshift(np.fft.ifft2(img1ft * np.conj(img2ft))))

    return xcorr

def _save_alignment_data(
    ref_image: FibsemImage,
    new_image: FibsemImage,
    bandpass: np.ndarray,
    xcorr: np.ndarray,
    ref_mask: np.ndarray = None,
    lowpass: float = None,
    highpass: float = None,
    sigma: float = None,
    xcorr_limit: float = None,
    use_rect_mask: bool = False,
    dx: float = None,
    dy: float = None,
    pixelsize_x: float = None,
    pixelsize_y: float = None,
    

):
    """Save alignment data to disk."""
    
    import pandas as pd
    import os
    from fibsem import config as cfg
    import tifffile as tff

    ts = utils.current_timestamp_v2()
    fname = os.path.join(cfg.DATA_CC_PATH, str(ts))

    # save fibsem images
    ref_image.save(fname + "_ref.tif")
    new_image.save(fname + "_new.tif")

    # convert to tiff , save
    tff.imwrite(fname + "_xcorr.tif", xcorr)
    tff.imwrite(fname + "_bandpass.tif", bandpass)
    if ref_mask is not None:
        tff.imwrite(fname + "_ref_mask.tif", ref_mask)

    info = {
        # "ref_image": ref_image, "new_image": new_image, "bandpass": bandpass, "xcorr": xcorr, "ref_mask": ref_mask,
        "lowpass": lowpass, "highpass": highpass, "sigma": sigma,
        "pixelsize_x": pixelsize_x, "pixelsize_y": pixelsize_y, 
        "use_rect_mask": use_rect_mask, "xcorr_limit": xcorr_limit, "ref_mask": ref_mask is not None,
        "dx": dx, "dy": dy, "fname": fname, "timestamp": ts }


    df = pd.DataFrame.from_dict(info, orient='index').T
    
    # save the dataframe to a csv file, append if the file already exists
    DATAFRAME_PATH = os.path.join(cfg.DATA_CC_PATH, "data.csv")
    if os.path.exists(DATAFRAME_PATH):
        df_tmp = pd.read_csv(DATAFRAME_PATH)
        df = pd.concat([df_tmp, df], axis=0, ignore_index=True)
    
    df.to_csv(DATAFRAME_PATH, index=False)

from fibsem.structures import ImageSettings
def _multi_step_alignment(microscope: FibsemMicroscope, image_settings: ImageSettings, 
    ref_image: FibsemImage, reduced_area: FibsemRectangle, alignment_current: float = None, steps:int = 3) -> None:
    
    # set alignment current
    if alignment_current is not None:
        initial_current = microscope.get("current", image_settings.beam_type)
        microscope.set("current", alignment_current, image_settings.beam_type)
        autocontrast = image_settings.autocontrast
        image_settings.autocontrast = False

    base_label = image_settings.filename
    for i in range(steps):
        image_settings.filename = f"{base_label}_{i:02d}"
        image_settings.beam_type = BeamType.ION
        beam_shift_alignment(microscope, image_settings, 
                                        ref_image=ref_image,
                                            reduced_area=reduced_area)
    # reset beam current
    if alignment_current is not None:
        microscope.set("current", initial_current, image_settings.beam_type)
        image_settings.autocontrast = autocontrast


def multi_step_alignment_v2(microscope: FibsemMicroscope, 
    ref_image: FibsemImage, beam_type: BeamType, 
    alignment_current: float = None, steps:int = 3) -> None:
    """Runs the beam shift alignment multiple times. Optionally sets the beam current before alignment."""
    # set alignment current
    if alignment_current is not None:
        initial_current = microscope.get("current", beam_type)
        microscope.set("current", alignment_current, beam_type)

    for i in range(steps):
        beam_shift_alignment_v2(microscope, ref_image=ref_image)
    
    # reset beam current
    if alignment_current is not None:
        microscope.set("current", initial_current, beam_type)


### From AutoLamella v1

# def realign(microscope, new_image, reference_image):
#     """Realign to reference image using beam shift.

#     Parameters
#     ----------
#     microscope : Autoscript microscope object
#     new_image : The most recent image acquired.
#         Must have the same dimensions and relative position as the reference.
#     reference_image : The reference image to align with.
#         Muast have the same dimensions and relative position as the new image
#     Returns
#     -------
#     microscope.beams.ion_beam.beam_shift.value
#         The current beam shift position (after any realignment)
#     """
#     from autoscript_core.common import ApplicationServerException

#     shift_in_meters = _calculate_beam_shift(new_image, reference_image)
#     try:
#         microscope.beams.ion_beam.beam_shift.value += shift_in_meters
#     except ApplicationServerException:
#         logging.warning(
#             "Cannot move beam shift beyond limits, "
#             "will continue with no beam shift applied."
#         )
#     return microscope.beams.ion_beam.beam_shift.value
    
# def _calculate_beam_shift(image_1, image_2):
#     """Cross correlation to find shift between two images.

#     Parameters
#     ----------
#     image_1 : AdornedImage
#         Original image to use as reference point.
#     image_2 : AdornedImage
#         Possibly shifted image to align with original.

#     Returns
#     -------
#     realspace_beam_shift
#         Beam shift in x, y format (meters), list of floats.

#     Raises
#     ------
#     ValueError
#         If images are not the same dimensions, raise a ValueError.
#     """
#     if image_1.data.shape != image_2.data.shape:
#         raise ValueError("Images must be the same shape for cross correlation.")
#     mask_image_1 = _mask_circular(image_1.data.shape)
#     mask_image_2 = _mask_rectangular(image_2.data.shape)
#     norm_image_1 = _normalize_image(image_1.data) * mask_image_1
#     norm_image_2 = _normalize_image(image_2.data) * mask_image_2
#     pixel_shift = _simple_register_translation(norm_image_2, norm_image_1)
#     # Autoscript y-axis has an inverted positive direction
#     pixel_shift[1] = -pixel_shift[1]
#     pixelsize_x = image_1.metadata.binary_result.pixel_size.x
#     realspace_beam_shift = pixel_shift * pixelsize_x
#     logging.info("pixel_shift calculated = {}".format(pixel_shift))
#     logging.info("realspace_beam_shift calculated = {}".format(realspace_beam_shift))
#     return realspace_beam_shift


# def _simple_register_translation(src_image, target_image, max_shift_mask=None):
#     """Calculate pixel shift between two input images.

#     This function runs with numpy or cupy for GPU acceleration.

#     Parameters
#     ----------
#     src_image : array
#         Reference image.
#     target_image : array
#         Image to register.  Must be same dimensionality as ``src_image``.
#     max_shift_mask : array
#         The fourier mask restricting the maximum allowable pixel shift.

#     Returns
#     -------
#     shifts : ndarray
#         Pixel shift in x, y order between target and source image.

#     References
#     ----------
#     scikit-image register_translation function in the skimage.feature module.
#     """
#     src_freq = np.fft.fftn(src_image)
#     target_freq = np.fft.fftn(target_image)
#     print('using bp mask')
#     bp_mask = _bandpass_mask(target_image.data.shape, target_image.data.shape[0] / 3, inner_radius=2, sigma=3)
#     bp_target_freq = bp_mask * target_freq

#     # Whole-pixel shift - Compute cross-correlation by an IFFT
#     shape = src_freq.shape
#     image_product = src_freq * bp_target_freq.conj()
#     cross_correlation = np.fft.ifftn(image_product)
#     # Locate maximum
#     maxima = np.unravel_index(
#         np.argmax(np.abs(cross_correlation)), cross_correlation.shape
#     )
#     midpoints = np.array([float(np.round(axis_size / 2)) for axis_size in shape])
#     shifts = np.array(maxima, dtype=np.float64)
#     shifts[shifts > midpoints] -= np.array(shape)[shifts > midpoints]
#     shifts = np.flip(shifts, axis=0).astype(int)  # x, y order
#     return shifts


# def _normalize_image(image, mask=None):
#     """Ensure the image mean is zero and the standard deviation is one.

#     Parameters
#     ----------
#     image : ndarray
#         The input image array.
#     mask : ndarray, optional
#         A mask image containing values between zero and one.
#         Dimensions must match the input image.

#     Returns
#     -------
#     ndarray
#         The normalized image.
#         The mean intensity is equal to zero and standard deviation equals one.
#     """
#     image = image - np.mean(image)
#     image = image / np.std(image)
#     if mask:
#         image = image * mask
#     return image


# def _mask_circular(image_shape, sigma=5.0, *, radius=None):
#     """Make a circular mask with soft edges for image normalization.

#     Parameters
#     ----------
#     image_shape : tuple
#         Shape of the original image array
#     sigma : float, optional
#         Sigma value (in pixels) for gaussian blur function, by default 5.
#     radius : int, optional
#         Radius of circle, by default None which will create a circle that fills
#         90% of the smallest image dimension.

#     Returns
#     -------
#     ndarray
#         Circular mask with soft edges in array matching the input image_shape
#     """
#     if radius is None:
#         # leave at least a 5% gap on each edge
#         radius = 0.45 * min(image_shape)
#     center = np.divide(image_shape, 2)
#     rr, cc = skimage.draw.disk(center, radius=radius, shape=image_shape)
#     mask = np.zeros(image_shape)
#     mask[rr, cc] = 1.0
#     mask = ndi.gaussian_filter(mask, sigma=sigma)
#     return mask


# def _mask_rectangular(image_shape, sigma=5.0, *, start=None, extent=None):
#     """Make a rectangular mask with soft edges for image normalization.

#     Parameters
#     ----------
#     image_shape : tuple
#         Shape of the original image array
#     sigma : float, optional
#         Sigma value (in pixels) for gaussian blur function, by default 5.
#     start : tuple, optional
#         Origin point of the rectangle, e.g., ([plane,] row, column).
#         Default start is 5% of the total image width and height.
#     extent : int, optional
#         The extent (size) of the drawn rectangle.
#         E.g., ([num_planes,] num_rows, num_cols).
#         Default is for the rectangle to cover 95% of the image width & height.

#     Returns
#     -------
#     ndarray
#         Rectangular mask with soft edges in array matching input image_shape.
#     """
#     if extent is None:
#         # leave at least a 5% gap on each edge
#         start = np.round(np.array(image_shape) * 0.05)
#         extent = np.round(np.array(image_shape) * 0.90)
#     rr, cc = skimage.draw.rectangle(start, extent=extent, shape=image_shape)
#     mask = np.zeros(image_shape)
#     mask[rr.astype(int), cc.astype(int)] = 1.0
#     mask = ndi.gaussian_filter(mask, sigma=sigma)
#     return mask


# def _bandpass_mask(image_shape, outer_radius, inner_radius=0, sigma=5):
#     """Create a fourier bandpass mask.

#     Parameters
#     ----------
#     image_shape : tuple
#         Shape of the original image array
#     outer_radius : int
#         Outer radius for bandpass filter array.
#     inner_radius : int, optional
#         Inner radius for bandpass filter array, by default 0
#     sigma : int, optional
#         Sigma value for edge blending, by default 5 pixels.

#     Returns
#     -------
#     _bandpass_mask : ndarray
#         The bandpass image mask.
#     """
#     _bandpass_mask = np.zeros(image_shape)
#     center = np.divide(image_shape, 2)
#     inner_circle_rr, inner_circle_cc = skimage.draw.disk(
#         center, inner_radius, shape=image_shape
#     )
#     outer_circle_rr, outer_circle_cc = skimage.draw.disk(
#         center, outer_radius, shape=image_shape
#     )
#     _bandpass_mask[outer_circle_rr, outer_circle_cc] = 1.0
#     _bandpass_mask[inner_circle_rr, inner_circle_cc] = 0.0
#     _bandpass_mask = ndi.gaussian_filter(_bandpass_mask, sigma)
#     _bandpass_mask = np.array(_bandpass_mask)
#     # fourier space origin should be in the corner
#     _bandpass_mask = np.roll(
#         _bandpass_mask, (np.array(image_shape) / 2).astype(int), axis=(0, 1)
#     )
#     return _bandpass_mask


import numpy as np
from fibsem.structures import Point, FibsemImage
from PIL import Image


def create_distance_map_px(w: int, h: int) -> np.ndarray:
    x = np.arange(0, w)
    y = np.arange(0, h)

    X, Y = np.meshgrid(x, y)
    distance = np.sqrt(((w / 2) - X) ** 2 + ((h / 2) - Y) ** 2)

    return distance


def measure_brightness(img: FibsemImage) -> float:

    return np.mean(img.data)


def rotate_image(image: FibsemImage):
    """Rotate the AdornedImage 180 degrees."""
    data = np.rot90(np.rot90(np.copy(image.data)))
    reference = FibsemImage(data=data, metadata=image.metadata)
    return reference


def normalise_image(img: FibsemImage) -> np.ndarray:
    """Normalise the image"""
    return (img.data - np.mean(img.data)) / np.std(img.data)


def cosine_stretch(img: FibsemImage, tilt_degrees: float):
    """Apply a cosine stretch to an image based on the relative tilt.

    This is required when aligning images with different tilts to ensure features are the same size.

    Args:
        img (AdornedImage): _description_
        tilt_degrees (float): _description_

    Returns:
        _type_: _description_
    """
    # note: do smaller version for negative tilt??

    tilt = np.deg2rad(tilt_degrees)

    shape = int(img.data.shape[0] / np.cos(tilt)), int(img.data.shape[1] / np.cos(tilt))

    # cosine stretch
    # larger
    resized_img = np.asarray(
        Image.fromarray(img.data).resize(size=(shape[1], shape[0]))
    )

    # crop centre out?
    c = Point(resized_img.shape[1] // 2, resized_img.shape[0] // 2)
    dy, dx = img.data.shape[0] // 2, img.data.shape[1] // 2
    scaled_img = resized_img[c.y - dy : c.y + dy, c.x - dx : c.x + dx]

    return FibsemImage(data=scaled_img, metadata=img.metadata)


def apply_image_mask(img: FibsemImage, mask: np.ndarray) -> np.ndarray:

    return normalise_image(img) * mask



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemPositionsWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(591, 505)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.comboBox_positions = QtWidgets.QComboBox(Form)
        self.comboBox_positions.setObjectName("comboBox_positions")
        self.gridLayout.addWidget(self.comboBox_positions, 2, 1, 1, 2)
        self.label = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.pushButton_remove_position = QtWidgets.QPushButton(Form)
        self.pushButton_remove_position.setObjectName("pushButton_remove_position")
        self.gridLayout.addWidget(self.pushButton_remove_position, 5, 0, 1, 1)
        self.pushButton_save_position = QtWidgets.QPushButton(Form)
        self.pushButton_save_position.setObjectName("pushButton_save_position")
        self.gridLayout.addWidget(self.pushButton_save_position, 4, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.lineEdit_position_name = QtWidgets.QLineEdit(Form)
        self.lineEdit_position_name.setObjectName("lineEdit_position_name")
        self.gridLayout.addWidget(self.lineEdit_position_name, 4, 1, 1, 2)
        self.pushButton_export = QtWidgets.QPushButton(Form)
        self.pushButton_export.setObjectName("pushButton_export")
        self.gridLayout.addWidget(self.pushButton_export, 6, 0, 1, 3)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 9, 1, 1, 1)
        self.pushButton_import = QtWidgets.QPushButton(Form)
        self.pushButton_import.setObjectName("pushButton_import")
        self.gridLayout.addWidget(self.pushButton_import, 7, 0, 1, 3)
        self.label_current_position = QtWidgets.QLabel(Form)
        self.label_current_position.setText("")
        self.label_current_position.setObjectName("label_current_position")
        self.gridLayout.addWidget(self.label_current_position, 3, 1, 1, 2)
        self.pushButton_go_to = QtWidgets.QPushButton(Form)
        self.pushButton_go_to.setObjectName("pushButton_go_to")
        self.gridLayout.addWidget(self.pushButton_go_to, 3, 0, 1, 1)
        self.label_minimap = QtWidgets.QLabel(Form)
        self.label_minimap.setText("")
        self.label_minimap.setObjectName("label_minimap")
        self.gridLayout.addWidget(self.label_minimap, 8, 0, 1, 3)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Positions "))
        self.pushButton_remove_position.setText(_translate("Form", "Remove Position"))
        self.pushButton_save_position.setText(_translate("Form", "Save position"))
        self.label_2.setText(_translate("Form", "Saved positions"))
        self.label_3.setText(_translate("Form", "All positions in mm and degrees"))
        self.pushButton_export.setText(_translate("Form", "Export positions"))
        self.pushButton_import.setText(_translate("Form", "Import Positions"))
        self.pushButton_go_to.setText(_translate("Form", "Go To"))
